#
# Copyright 2018-2022 Elyra Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

"""Tests for elyra-metadata application"""
import json
import os
import shutil
from tempfile import mkdtemp
from tempfile import TemporaryDirectory
from typing import List
from typing import Optional

import pytest

from elyra.metadata.manager import MetadataManager
from elyra.metadata.metadata import Metadata
from elyra.metadata.schema import METADATA_TEST_SCHEMASPACE
from elyra.metadata.schema import METADATA_TEST_SCHEMASPACE_ID
from elyra.tests.metadata.test_utils import all_of_json
from elyra.tests.metadata.test_utils import another_metadata_json
from elyra.tests.metadata.test_utils import create_json_file
from elyra.tests.metadata.test_utils import invalid_metadata_json
from elyra.tests.metadata.test_utils import invalid_no_display_name_json
from elyra.tests.metadata.test_utils import invalid_schema_name_json
from elyra.tests.metadata.test_utils import one_of_json
from elyra.tests.metadata.test_utils import PropertyTester
from elyra.tests.metadata.test_utils import valid_display_name_json
from elyra.tests.metadata.test_utils import valid_metadata2_json
from elyra.tests.metadata.test_utils import valid_metadata_json

os.environ["METADATA_TESTING"] = "1"  # Enable metadata-tests schemaspace


@pytest.fixture()
def mock_data_dir():
    runtime_dir = mkdtemp(prefix="runtime_")
    orig_data_dir = os.environ.get("JUPYTER_DATA_DIR")
    os.environ["JUPYTER_DATA_DIR"] = runtime_dir
    yield runtime_dir  # provide the fixture value
    shutil.rmtree(runtime_dir)
    if orig_data_dir:
        os.environ["JUPYTER_DATA_DIR"] = orig_data_dir
    else:
        os.environ.pop("JUPYTER_DATA_DIR")


def test_no_opts(script_runner):
    ret = script_runner.run("elyra-metadata")
    assert ret.success is False
    message = (
        "No subcommand specified.  One of: "
        "['list', 'create', 'update', 'install', 'remove', 'migrate', 'export', 'import'] "
        "must be specified."
    )
    assert message in ret.stdout


def test_bad_subcommand(script_runner):
    ret = script_runner.run("elyra-metadata", "bogus-subcommand")
    assert ret.success is False
    assert (
        "Subcommand 'bogus-subcommand' is invalid.  One of: "
        "['list', 'create', 'update', 'install', 'remove', 'migrate', 'export', 'import'] "
        "must be specified." in ret.stdout
    )


def test_install_bad_argument(script_runner):
    ret = script_runner.run("elyra-metadata", "install", "--bogus-argument")
    assert ret.success is False
    assert "Subcommand '--bogus-argument' is invalid." in ret.stdout
    assert f"Install a metadata instance into schemaspace '{METADATA_TEST_SCHEMASPACE}'." in ret.stdout


def test_install_bad_schemaspace(script_runner):
    ret = script_runner.run("elyra-metadata", "install", "bogus-schemaspace")
    assert ret.success is False
    assert "Subcommand 'bogus-schemaspace' is invalid." in ret.stdout
    assert f"Install a metadata instance into schemaspace '{METADATA_TEST_SCHEMASPACE}'." in ret.stdout


def test_install_help(script_runner):
    ret = script_runner.run("elyra-metadata", "install", METADATA_TEST_SCHEMASPACE, "--help")
    assert ret.success is False
    assert f"Install a metadata instance into schemaspace '{METADATA_TEST_SCHEMASPACE}'." in ret.stdout


def test_install_no_schema_single(script_runner, mock_data_dir):
    # Use the runtime-images schemaspace since that is most likely to always be a single-schema schemaspace.
    # Note: this test will break if it ever supports multiple.
    ret = script_runner.run("elyra-metadata", "install", "runtime-images")
    assert ret.success is False
    assert "ERROR: '--display_name' is a required parameter." in ret.stdout


def test_install_no_schema_multiple(script_runner, mock_data_dir):
    ret = script_runner.run("elyra-metadata", "install", METADATA_TEST_SCHEMASPACE)
    assert ret.success is False
    # Since order in dictionaries, where the one-of list is derived, can be random, just check up to the
    # first known difference in the schema names.
    assert (
        "ERROR: '--schema_name' is a required parameter and must be one of the "
        "following values: ['metadata-test" in ret.stdout
    )


def test_install_bad_schema_multiple(script_runner, mock_data_dir):
    ret = script_runner.run("elyra-metadata", "install", METADATA_TEST_SCHEMASPACE, "--schema_name=metadata-foo")
    assert ret.success is False
    assert "ERROR: Parameter '--schema_name' requires one of the " "following values: ['metadata-test" in ret.stdout


def test_install_no_name(script_runner, mock_data_dir):
    ret = script_runner.run("elyra-metadata", "install", METADATA_TEST_SCHEMASPACE, "--schema_name=metadata-test")
    assert ret.success is False
    assert "ERROR: '--display_name' is a required parameter." in ret.stdout


def test_install_complex_usage(script_runner, mock_data_dir):
    ret = script_runner.run("elyra-metadata", "install", METADATA_TEST_SCHEMASPACE, "--schema_name=metadata-test")
    assert ret.success is False
    assert "Note: The following properties in this schema contain JSON keywords that are not supported" in ret.stdout
    assert "*** References unsupported keywords: {'oneOf'}" in ret.stdout
    assert "*** References unsupported keywords: {'allOf'}" in ret.stdout
    assert "*** References unsupported keywords: {'$ref'}" in ret.stdout


def test_install_only_display_name(script_runner, mock_data_dir):
    metadata_display_name = "1 teste 'rápido'"
    metadata_name = "a_1_teste_rpido"

    ret = script_runner.run(
        "elyra-metadata",
        "install",
        METADATA_TEST_SCHEMASPACE,
        "--schema_name=metadata-test",
        f"--display_name={metadata_display_name}",
        "--required_test=required_value",
    )
    assert ret.success is True
    assert f"Metadata instance '{metadata_name}' for schema 'metadata-test' has been written to:" in ret.stdout

    # Ensure it can be fetched by name...
    metadata_manager = MetadataManager(schemaspace=METADATA_TEST_SCHEMASPACE_ID)
    resource = metadata_manager.get(metadata_name)
    assert resource.display_name == metadata_display_name


def test_install_invalid_name(script_runner, mock_data_dir):
    ret = script_runner.run(
        "elyra-metadata",
        "install",
        METADATA_TEST_SCHEMASPACE,
        "--schema_name=metadata-test",
        "--name=UPPER_CASE_NOT_ALLOWED",
        "--display_name=display_name",
        "--required_test=required_value",
    )
    assert ret.success is False
    assert "The following exception occurred saving metadata instance for schema 'metadata-test'" in ret.stdout
    assert "Name of metadata must be lowercase alphanumeric" in ret.stdout


def test_install_simple(script_runner, mock_data_dir):
    expected_file = os.path.join(
        mock_data_dir, "metadata", METADATA_TEST_SCHEMASPACE, "test-metadata_42_valid-name.json"
    )
    # Cleanup from any potential previous failures
    if os.path.exists(expected_file):
        os.remove(expected_file)

    ret = script_runner.run(
        "elyra-metadata",
        "install",
        METADATA_TEST_SCHEMASPACE,
        "--schema_name=metadata-test",
        "--name=test-metadata_42_valid-name",
        "--display_name=display_name",
        "--required_test=required_value",
    )
    assert ret.success
    assert "Metadata instance 'test-metadata_42_valid-name' for schema 'metadata-test' has been written" in ret.stdout

    assert os.path.isdir(os.path.join(mock_data_dir, "metadata", METADATA_TEST_SCHEMASPACE))
    assert os.path.isfile(expected_file)

    with open(expected_file, "r") as fd:
        instance_json = json.load(fd)
        assert instance_json["schema_name"] == "metadata-test"
        assert instance_json["display_name"] == "display_name"
        assert instance_json["metadata"]["required_test"] == "required_value"
        assert instance_json["metadata"]["number_default_test"] == 42  # defaults will always persist


@pytest.mark.parametrize("option_style", ["equals", "sans-equals", "missing"])
def test_create_from_file(script_runner, mock_data_dir, option_style):
    content = create_json_file(mock_data_dir, "valid.json", valid_metadata_json)

    argv: List[str] = [
        "elyra-metadata",
        "install",
        METADATA_TEST_SCHEMASPACE,
    ]

    if option_style == "equals":
        argv.append(f"--file={content}")
    elif option_style == "sans-equals":
        argv.append("--file")
        argv.append(f"{content}")
    else:  # missing
        argv.append("--file")

    ret = script_runner.run(*argv)

    if option_style == "missing":
        assert ret.success is False
        assert (
            "ERROR: Parameter '--file' requires a file with format JSON and no value was provided.  "
            "Try again with an appropriate value." in ret.stdout
        )
    else:  # success expected
        assert ret.success
        assert "Metadata instance 'valid' for schema 'metadata-test' has been written" in ret.stdout


@pytest.mark.parametrize("option_style", ["equals", "sans-equals", "missing"])
def test_create_from_json(script_runner, mock_data_dir, option_style):

    content = json.dumps(valid_metadata_json)

    argv: List[str] = [
        "elyra-metadata",
        "install",
        METADATA_TEST_SCHEMASPACE,
    ]

    if option_style == "equals":
        argv.append(f"--json={content}")
    elif option_style == "sans-equals":
        argv.append("--json")
        argv.append(f"{content}")
    else:  # missing
        argv.append("--json")

    ret = script_runner.run(*argv)

    if option_style == "missing":
        assert ret.success is False
        assert (
            "ERROR: Parameter '--json' requires a value with format JSON and no value was provided.  "
            "Try again with an appropriate value." in ret.stdout
        )
    else:  # success expected
        assert ret.success
        assert "Metadata instance 'valid_metadata_instance' for schema 'metadata-test' has been written" in ret.stdout


def test_install_and_replace(script_runner, mock_data_dir):
    expected_file = os.path.join(
        mock_data_dir, "metadata", METADATA_TEST_SCHEMASPACE, "test-metadata_42_valid-name.json"
    )
    # Cleanup from any potential previous failures
    if os.path.exists(expected_file):
        os.remove(expected_file)

    # Attempt replace before schemaspace exists and ensure appropriate error message
    ret = script_runner.run(
        "elyra-metadata",
        "install",
        METADATA_TEST_SCHEMASPACE,
        "--schema_name=metadata-test",
        "--name=test-metadata_42_valid-name",
        "--display_name=display_name",
        "--required_test=required_value",
        "--replace",
    )
    assert ret.success is False
    assert (
        "No such instance named 'test-metadata_42_valid-name' was found in the metadata-tests schemaspace."
        in ret.stdout
    )

    ret = script_runner.run(
        "elyra-metadata",
        "install",
        METADATA_TEST_SCHEMASPACE,
        "--schema_name=metadata-test",
        "--name=test-metadata_42_valid-name",
        "--display_name=display_name",
        "--required_test=required_value",
        "--number_default_test=24",
    )
    assert ret.success
    assert "Metadata instance 'test-metadata_42_valid-name' for schema 'metadata-test' has been written" in ret.stdout
    assert expected_file in ret.stdout
    assert os.path.isdir(os.path.join(mock_data_dir, "metadata", METADATA_TEST_SCHEMASPACE))
    assert os.path.isfile(expected_file)

    with open(expected_file, "r") as fd:
        instance_json = json.load(fd)
        assert instance_json["metadata"]["number_default_test"] == 24  # ensure CLI value is used over default

    # Re-attempt w/o replace flag - failure expected
    ret = script_runner.run(
        "elyra-metadata",
        "install",
        METADATA_TEST_SCHEMASPACE,
        "--schema_name=metadata-test",
        "--name=test-metadata_42_valid-name",
        "--display_name=display_name",
        "--required_test=required_value",
    )
    assert ret.success is False
    assert (
        "An instance named 'test-metadata_42_valid-name' already exists in the metadata-tests "
        "schemaspace" in ret.stderr
    )

    # Re-attempt with replace flag but without --name - failure expected
    ret = script_runner.run(
        "elyra-metadata",
        "install",
        METADATA_TEST_SCHEMASPACE,
        "--schema_name=metadata-test",
        "--display_name=display_name",
        "--required_test=required_value",
        "--replace",
    )
    assert ret.success is False
    assert "The 'name' parameter requires a value" in ret.stdout

    # Re-attempt with replace flag - success expected
    ret = script_runner.run(
        "elyra-metadata",
        "install",
        METADATA_TEST_SCHEMASPACE,
        "--schema_name=metadata-test",
        "--name=test-metadata_42_valid-name",
        "--display_name=display_name",
        "--required_test=required_value",
        "--replace",
    )
    assert ret.success
    assert "Metadata instance 'test-metadata_42_valid-name' for schema 'metadata-test' has been written" in ret.stdout

    assert os.path.isdir(os.path.join(mock_data_dir, "metadata", METADATA_TEST_SCHEMASPACE))
    assert os.path.isfile(expected_file)

    with open(expected_file, "r") as fd:
        instance_json = json.load(fd)
        assert instance_json["schema_name"] == "metadata-test"
        assert instance_json["display_name"] == "display_name"
        assert instance_json["metadata"]["required_test"] == "required_value"
        assert instance_json["metadata"]["number_default_test"] == 24  # ensure original value is used over default


@pytest.mark.parametrize("complex_keyword", ["defs", "oneOf", "allOf"])
def test_install_and_replace_complex(script_runner, mock_data_dir, complex_keyword):

    # create and use deep copies of the global one_of_json and one_of_json
    # to avoid side effects
    one_of_json_cp = json.loads(json.dumps(one_of_json))
    all_of_json_cp = json.loads(json.dumps(all_of_json))

    test_file: Optional[str] = None
    name: str = f"test-complex-{complex_keyword}".lower()

    if complex_keyword == "defs":
        option = "--json"
        value = '{ "defs_test": 42 }'

    elif complex_keyword == "oneOf":
        option = "--file"
        # Build the file...
        test_file = os.path.join(mock_data_dir, f"{complex_keyword}.json")
        with open(test_file, mode="w") as one_of_fd:
            json.dump(one_of_json_cp, one_of_fd)
        value = test_file
    else:  # allOf
        option = "--allOf_test"  # Use "ovp-from-file" approach
        # Build the file...
        test_file = os.path.join(mock_data_dir, f"{complex_keyword}.json")
        with open(test_file, mode="w") as all_of_fd:
            json.dump(all_of_json_cp, all_of_fd)
        value = test_file

    expected_file = os.path.join(mock_data_dir, "metadata", METADATA_TEST_SCHEMASPACE, f"{name}.json")
    # Cleanup from any potential previous failures (should be rare)
    if os.path.exists(expected_file):
        os.remove(expected_file)

    ret = script_runner.run(
        "elyra-metadata",
        "install",
        METADATA_TEST_SCHEMASPACE,
        "--schema_name=metadata-test",
        f"--name={name}",
        f"--display_name=Test Complex {complex_keyword}",
        "--required_test=required_value",
        f"{option}={value}",
    )
    assert ret.success
    assert f"Metadata instance '{name}' for schema 'metadata-test' has been written" in ret.stdout
    assert expected_file in ret.stdout
    assert os.path.exists(expected_file)

    with open(expected_file) as fd:
        json_results = json.load(fd)

    # Verify common stuff
    assert json_results["display_name"] == f"Test Complex {complex_keyword}"
    assert json_results["metadata"]["required_test"] == "required_value"

    # Verify result and prepare for replace...
    if complex_keyword == "defs":
        assert json_results["metadata"]["defs_test"] == 42
        value = '{ "defs_test": 24 }'
    elif complex_keyword == "oneOf":
        assert json_results["metadata"]["oneOf_test"]["obj_switch"] == "obj2"
        assert json_results["metadata"]["oneOf_test"]["obj2_prop1"] == 42
        one_of_json_cp["metadata"]["oneOf_test"]["obj2_prop1"] = 24
        with open(test_file, mode="w+") as one_of_fd:
            json.dump(one_of_json_cp, one_of_fd)
    elif complex_keyword == "allOf":
        assert len(json_results["metadata"]["allOf_test"]) == 9
        assert json_results["metadata"]["allOf_test"]["obj1_switch"] == "obj1"
        assert json_results["metadata"]["allOf_test"]["obj1_prop1"] == "allOf-test-val1"
        assert json_results["metadata"]["allOf_test"]["obj1_prop2"] == "allOf-test-val2"
        all_of_json_cp["obj1_prop1"] = "allOf-test-val1-replace"
        assert json_results["metadata"]["allOf_test"]["obj2_switch"] == "obj2"
        assert json_results["metadata"]["allOf_test"]["obj2_prop1"] == 42
        assert json_results["metadata"]["allOf_test"]["obj2_prop2"] == 24
        all_of_json_cp["obj2_prop1"] = 24
        assert json_results["metadata"]["allOf_test"]["obj3_switch"] == "obj3"
        assert json_results["metadata"]["allOf_test"]["obj3_prop1"] == 42.7
        assert json_results["metadata"]["allOf_test"]["obj3_prop2"] is True
        all_of_json_cp["obj3_prop1"] = 7.24

        with open(test_file, mode="w+") as all_of_fd:
            json.dump(all_of_json_cp, all_of_fd)

    # Replace the previously-created instance
    ret = script_runner.run(
        "elyra-metadata",
        "install",
        METADATA_TEST_SCHEMASPACE,
        "--schema_name=metadata-test",
        f"--name={name}",
        f"--display_name=Test Complex {complex_keyword}2",
        "--required_test=required_value",
        f"{option}={value}",
        "--replace",
    )
    assert ret.success
    assert f"Metadata instance '{name}' for schema 'metadata-test' has been written" in ret.stdout
    assert expected_file in ret.stdout
    assert os.path.exists(expected_file)

    with open(expected_file) as fd:
        json_results = json.load(fd)

    # Verify common stuff
    assert json_results["display_name"] == f"Test Complex {complex_keyword}2"
    assert json_results["metadata"]["required_test"] == "required_value"

    # Verify result following replace...
    if complex_keyword == "defs":
        assert json_results["metadata"]["defs_test"] == 24
    elif complex_keyword == "oneOf":
        assert json_results["metadata"]["oneOf_test"]["obj_switch"] == "obj2"
        assert json_results["metadata"]["oneOf_test"]["obj2_prop1"] == 24
        assert json_results["metadata"]["oneOf_test"]["obj2_prop2"] == 24
    elif complex_keyword == "allOf":
        assert len(json_results["metadata"]["allOf_test"]) == 9
        assert json_results["metadata"]["allOf_test"]["obj1_prop1"] == "allOf-test-val1-replace"
        assert json_results["metadata"]["allOf_test"]["obj1_prop2"] == "allOf-test-val2"
        assert json_results["metadata"]["allOf_test"]["obj2_prop1"] == 24
        assert json_results["metadata"]["allOf_test"]["obj2_prop2"] == 24
        assert json_results["metadata"]["allOf_test"]["obj3_prop1"] == 7.24
        assert json_results["metadata"]["allOf_test"]["obj3_prop2"] is True


# ---------- begin of 'create' command tests


def test_create_bad_argument(script_runner):
    ret = script_runner.run("elyra-metadata", "create", "--bogus-argument")
    assert ret.success is False
    assert "Subcommand '--bogus-argument' is invalid." in ret.stdout
    assert f"Create a metadata instance in schemaspace '{METADATA_TEST_SCHEMASPACE}'." in ret.stdout


def test_create_bad_schemaspace(script_runner):
    ret = script_runner.run("elyra-metadata", "create", "bogus-schemaspace")
    assert ret.success is False
    assert "Subcommand 'bogus-schemaspace' is invalid." in ret.stdout
    assert f"Create a metadata instance in schemaspace '{METADATA_TEST_SCHEMASPACE}'." in ret.stdout


def test_create_help(script_runner):
    ret = script_runner.run("elyra-metadata", "create", METADATA_TEST_SCHEMASPACE, "--help")
    assert ret.success is False
    assert f"Create a metadata instance in schemaspace '{METADATA_TEST_SCHEMASPACE}'." in ret.stdout


def test_create_no_schema_single(script_runner, mock_data_dir):
    # Use the runtime-images schemaspace since that is most likely to always be a single-schema schemaspace.
    # Note: this test will break if it ever supports multiple.
    ret = script_runner.run("elyra-metadata", "create", "runtime-images")
    assert ret.success is False
    assert "ERROR: '--display_name' is a required parameter." in ret.stdout


def test_create_no_schema_multiple(script_runner, mock_data_dir):
    ret = script_runner.run("elyra-metadata", "create", METADATA_TEST_SCHEMASPACE)
    assert ret.success is False
    # Since order in dictionaries, where the one-of list is derived, can be random, just check up to the
    # first known difference in the schema names.
    assert (
        "ERROR: '--schema_name' is a required parameter and must be one of the "
        "following values: ['metadata-test" in ret.stdout
    )


def test_create_bad_schema_multiple(script_runner, mock_data_dir):
    ret = script_runner.run("elyra-metadata", "create", METADATA_TEST_SCHEMASPACE, "--schema_name=metadata-foo")
    assert ret.success is False
    assert "ERROR: Parameter '--schema_name' requires one of the " "following values: ['metadata-test" in ret.stdout


def test_create_no_name(script_runner, mock_data_dir):
    ret = script_runner.run("elyra-metadata", "create", METADATA_TEST_SCHEMASPACE, "--schema_name=metadata-test")
    assert ret.success is False
    assert "ERROR: '--display_name' is a required parameter." in ret.stdout


def test_create_complex_usage(script_runner, mock_data_dir):
    ret = script_runner.run("elyra-metadata", "create", METADATA_TEST_SCHEMASPACE, "--schema_name=metadata-test")
    assert ret.success is False
    assert "Note: The following properties in this schema contain JSON keywords that are not supported" in ret.stdout
    assert "*** References unsupported keywords: {'oneOf'}" in ret.stdout
    assert "*** References unsupported keywords: {'allOf'}" in ret.stdout
    assert "*** References unsupported keywords: {'$ref'}" in ret.stdout


def test_create_only_display_name(script_runner, mock_data_dir):
    metadata_display_name = "1 teste 'rápido'"
    metadata_name = "a_1_teste_rpido"

    ret = script_runner.run(
        "elyra-metadata",
        "create",
        METADATA_TEST_SCHEMASPACE,
        "--schema_name=metadata-test",
        f"--display_name={metadata_display_name}",
        "--required_test=required_value",
    )
    assert ret.success is True
    assert f"Metadata instance '{metadata_name}' for schema 'metadata-test' has been written to:" in ret.stdout

    # Ensure it can be fetched by name...
    metadata_manager = MetadataManager(schemaspace=METADATA_TEST_SCHEMASPACE_ID)
    resource = metadata_manager.get(metadata_name)
    assert resource.display_name == metadata_display_name


def test_create_invalid_name(script_runner, mock_data_dir):
    ret = script_runner.run(
        "elyra-metadata",
        "create",
        METADATA_TEST_SCHEMASPACE,
        "--schema_name=metadata-test",
        "--name=UPPER_CASE_NOT_ALLOWED",
        "--display_name=display_name",
        "--required_test=required_value",
    )
    assert ret.success is False
    assert "The following exception occurred saving metadata instance for schema 'metadata-test'" in ret.stdout
    assert "Name of metadata must be lowercase alphanumeric" in ret.stdout


def test_create_simple(script_runner, mock_data_dir):
    expected_file = os.path.join(
        mock_data_dir, "metadata", METADATA_TEST_SCHEMASPACE, "test-metadata_42_valid-name.json"
    )
    # Cleanup from any potential previous failures
    if os.path.exists(expected_file):
        os.remove(expected_file)

    ret = script_runner.run(
        "elyra-metadata",
        "create",
        METADATA_TEST_SCHEMASPACE,
        "--schema_name=metadata-test",
        "--name=test-metadata_42_valid-name",
        "--display_name=display_name",
        "--required_test=required_value",
    )
    assert ret.success
    assert "Metadata instance 'test-metadata_42_valid-name' for schema 'metadata-test' has been written" in ret.stdout

    assert os.path.isdir(os.path.join(mock_data_dir, "metadata", METADATA_TEST_SCHEMASPACE))
    assert os.path.isfile(expected_file)

    with open(expected_file, "r") as fd:
        instance_json = json.load(fd)
        assert instance_json["schema_name"] == "metadata-test"
        assert instance_json["display_name"] == "display_name"
        assert instance_json["metadata"]["required_test"] == "required_value"
        assert instance_json["metadata"]["number_default_test"] == 42  # defaults will always persist


def test_create_existing(script_runner, mock_data_dir):
    expected_file = os.path.join(
        mock_data_dir, "metadata", METADATA_TEST_SCHEMASPACE, "test-metadata_42_valid-name.json"
    )
    # Cleanup from any potential previous failures
    if os.path.exists(expected_file):
        os.remove(expected_file)

    ret = script_runner.run(
        "elyra-metadata",
        "create",
        METADATA_TEST_SCHEMASPACE,
        "--schema_name=metadata-test",
        "--name=test-metadata_42_valid-name",
        "--display_name=display_name",
        "--required_test=required_value",
        "--number_default_test=24",
    )
    assert ret.success
    assert "Metadata instance 'test-metadata_42_valid-name' for schema 'metadata-test' has been written" in ret.stdout
    assert expected_file in ret.stdout
    assert os.path.isdir(os.path.join(mock_data_dir, "metadata", METADATA_TEST_SCHEMASPACE))
    assert os.path.isfile(expected_file)

    with open(expected_file, "r") as fd:
        instance_json = json.load(fd)
        assert instance_json["metadata"]["number_default_test"] == 24  # ensure CLI value is used over default

    # Re-attempt create - failure expected
    ret = script_runner.run(
        "elyra-metadata",
        "create",
        METADATA_TEST_SCHEMASPACE,
        "--schema_name=metadata-test",
        "--name=test-metadata_42_valid-name",
        "--display_name=display_name",
        "--required_test=required_value",
    )

    assert ret.success is False
    assert (
        "An instance named 'test-metadata_42_valid-name' already exists in the metadata-tests "
        "schemaspace" in ret.stderr
    )


def test_create_complex(script_runner, mock_data_dir):

    complex_keyword = "defs"
    name: str = f"test-complex-{complex_keyword}".lower()

    option = "--json"
    value = '{ "defs_test": 42 }'

    expected_file = os.path.join(mock_data_dir, "metadata", METADATA_TEST_SCHEMASPACE, f"{name}.json")
    # Cleanup from any potential previous failures (should be rare)
    if os.path.exists(expected_file):
        os.remove(expected_file)

    ret = script_runner.run(
        "elyra-metadata",
        "create",
        METADATA_TEST_SCHEMASPACE,
        "--schema_name=metadata-test",
        f"--name={name}",
        f"--display_name=Test Complex {complex_keyword}",
        "--required_test=required_value",
        f"{option}",  # don't use '=' between option and value
        f"{value}",
    )
    assert ret.success
    assert f"Metadata instance '{name}' for schema 'metadata-test' has been written" in ret.stdout
    assert expected_file in ret.stdout
    assert os.path.exists(expected_file)

    with open(expected_file) as fd:
        json_results = json.load(fd)

    # Verify common stuff
    assert json_results["display_name"] == f"Test Complex {complex_keyword}"
    assert json_results["metadata"]["required_test"] == "required_value"
    assert json_results["metadata"]["defs_test"] == 42


# ---------- end of 'create' command tests
#
# ---------- begin of 'update' command tests


def test_update_bad_argument(script_runner):
    ret = script_runner.run("elyra-metadata", "update", "--bogus-argument")
    assert ret.success is False
    assert "Subcommand '--bogus-argument' is invalid." in ret.stdout
    assert f"Update a metadata instance in schemaspace '{METADATA_TEST_SCHEMASPACE}'." in ret.stdout


def test_update_bad_schemaspace(script_runner):
    ret = script_runner.run("elyra-metadata", "update", "bogus-schemaspace")
    assert ret.success is False
    assert "Subcommand 'bogus-schemaspace' is invalid." in ret.stdout
    assert f"Update a metadata instance in schemaspace '{METADATA_TEST_SCHEMASPACE}'." in ret.stdout


def test_update_help(script_runner):
    ret = script_runner.run("elyra-metadata", "update", METADATA_TEST_SCHEMASPACE, "--help")
    assert ret.success is False
    assert f"Update a metadata instance in schemaspace '{METADATA_TEST_SCHEMASPACE}'." in ret.stdout


def test_update_no_schema_single(script_runner, mock_data_dir):
    # Use the runtime-images schemaspace since that is most likely to always be a single-schema schemaspace.
    # Note: this test will break if it ever supports multiple.
    ret = script_runner.run("elyra-metadata", "update", "runtime-images")
    assert ret.success is False
    assert (
        "The following exception occurred saving metadata instance for schema 'runtime-image': "
        "The 'name' parameter requires a value." in ret.stdout
    )


def test_update_no_schema_multiple(script_runner, mock_data_dir):
    ret = script_runner.run("elyra-metadata", "update", METADATA_TEST_SCHEMASPACE)
    assert ret.success is False
    # Since order in dictionaries, where the one-of list is derived, can be random, just check up to the
    # first known difference in the schema names.
    assert (
        "ERROR: '--schema_name' is a required parameter and must be one of the "
        "following values: ['metadata-test" in ret.stdout
    )


def test_update_bad_schema_multiple(script_runner, mock_data_dir):
    ret = script_runner.run("elyra-metadata", "update", METADATA_TEST_SCHEMASPACE, "--schema_name=metadata-foo")
    assert ret.success is False
    assert "ERROR: Parameter '--schema_name' requires one of the " "following values: ['metadata-test" in ret.stdout


def test_update_no_name(script_runner, mock_data_dir):
    ret = script_runner.run("elyra-metadata", "update", METADATA_TEST_SCHEMASPACE, "--schema_name=metadata-test")
    assert ret.success is False
    assert (
        "The following exception occurred saving metadata instance for schema 'metadata-test': "
        "The 'name' parameter requires a value." in ret.stdout
    )


def test_update_no_instance(script_runner, mock_data_dir):
    """Attempt replace before instance exists"""
    ret = script_runner.run(
        "elyra-metadata",
        "update",
        METADATA_TEST_SCHEMASPACE,
        "--schema_name=metadata-test",
        "--name=test-metadata_42_valid-name",
        "--display_name=display_name",
        "--required_test=required_value",
    )
    assert ret.success is False
    assert (
        "No such instance named 'test-metadata_42_valid-name' was found in the metadata-tests schemaspace."
        in ret.stdout
    )


def test_update_simple(script_runner, mock_data_dir):
    expected_file = os.path.join(
        mock_data_dir, "metadata", METADATA_TEST_SCHEMASPACE, "test-metadata_42_valid-name.json"
    )
    # Cleanup from any potential previous failures
    if os.path.exists(expected_file):
        os.remove(expected_file)

    # create an instance
    ret = script_runner.run(
        "elyra-metadata",
        "create",
        METADATA_TEST_SCHEMASPACE,
        "--schema_name=metadata-test",
        "--name=test-metadata_42_valid-name",
        "--display_name=display_name",
        "--required_test=required_value",
    )
    assert ret.success
    assert "Metadata instance 'test-metadata_42_valid-name' for schema 'metadata-test' has been written" in ret.stdout

    # update instance
    ret = script_runner.run(
        "elyra-metadata",
        "update",
        METADATA_TEST_SCHEMASPACE,
        "--schema_name=metadata-test",
        "--name=test-metadata_42_valid-name",
        "--display_name=display_name",
        "--required_test=updated_required_value",
    )
    assert ret.success
    assert "Metadata instance 'test-metadata_42_valid-name' for schema 'metadata-test' has been written" in ret.stdout

    assert os.path.isdir(os.path.join(mock_data_dir, "metadata", METADATA_TEST_SCHEMASPACE))
    assert os.path.isfile(expected_file)

    with open(expected_file, "r") as fd:
        instance_json = json.load(fd)
        assert instance_json["schema_name"] == "metadata-test"
        assert instance_json["display_name"] == "display_name"
        assert instance_json["metadata"]["required_test"] == "updated_required_value"
        assert instance_json["metadata"]["number_default_test"] == 42  # defaults will always persist


@pytest.mark.parametrize("complex_keyword", ["defs", "oneOf", "allOf"])
def test_update_complex(script_runner, mock_data_dir, complex_keyword):

    # create and use deep copies of the global one_of_json and one_of_json
    # to avoid side effects
    one_of_json_cp = json.loads(json.dumps(one_of_json))
    all_of_json_cp = json.loads(json.dumps(all_of_json))

    test_file: Optional[str] = None
    name: str = f"test-complex-{complex_keyword}".lower()

    if complex_keyword == "defs":
        option = "--json"
        value = '{ "defs_test": 42 }'
    elif complex_keyword == "oneOf":
        option = "--file"
        # Build the file...
        test_file = os.path.join(mock_data_dir, f"{complex_keyword}.json")
        with open(test_file, mode="w") as one_of_fd:
            json.dump(one_of_json_cp, one_of_fd)
        value = test_file
    else:  # allOf
        option = "--allOf_test"  # Use "ovp-from-file" approach
        # Build the file...
        test_file = os.path.join(mock_data_dir, f"{complex_keyword}.json")
        with open(test_file, mode="w") as all_of_fd:
            json.dump(all_of_json_cp, all_of_fd)
        value = test_file

    expected_file = os.path.join(mock_data_dir, "metadata", METADATA_TEST_SCHEMASPACE, f"{name}.json")
    # Cleanup from any potential previous failures (should be rare)
    if os.path.exists(expected_file):
        os.remove(expected_file)

    # create instance
    ret = script_runner.run(
        "elyra-metadata",
        "create",
        METADATA_TEST_SCHEMASPACE,
        "--schema_name=metadata-test",
        f"--name={name}",
        f"--display_name=Test Complex {complex_keyword}",
        "--required_test=required_value",
        f"{option}={value}",
    )
    assert ret.success
    assert f"Metadata instance '{name}' for schema 'metadata-test' has been written" in ret.stdout
    assert expected_file in ret.stdout
    assert os.path.exists(expected_file)

    with open(expected_file) as fd:
        json_results = json.load(fd)

    # Verify common stuff
    assert json_results["display_name"] == f"Test Complex {complex_keyword}"
    assert json_results["metadata"]["required_test"] == "required_value"

    # Verify result and prepare for replace...
    if complex_keyword == "defs":
        assert json_results["metadata"]["defs_test"] == 42
        value = '{ "defs_test": 24 }'
    elif complex_keyword == "oneOf":
        assert json_results["metadata"]["oneOf_test"]["obj_switch"] == "obj2"
        assert json_results["metadata"]["oneOf_test"]["obj2_prop1"] == 42, f"--> {json_results}"
        one_of_json_cp["metadata"]["oneOf_test"]["obj2_prop1"] = 24
        with open(test_file, mode="w+") as one_of_fd:
            json.dump(one_of_json_cp, one_of_fd)
    elif complex_keyword == "allOf":
        assert len(json_results["metadata"]["allOf_test"]) == 9
        assert json_results["metadata"]["allOf_test"]["obj1_switch"] == "obj1"
        assert json_results["metadata"]["allOf_test"]["obj1_prop1"] == "allOf-test-val1"
        assert json_results["metadata"]["allOf_test"]["obj1_prop2"] == "allOf-test-val2"
        all_of_json_cp["obj1_prop1"] = "allOf-test-val1-replace"
        assert json_results["metadata"]["allOf_test"]["obj2_switch"] == "obj2"
        assert json_results["metadata"]["allOf_test"]["obj2_prop1"] == 42
        assert json_results["metadata"]["allOf_test"]["obj2_prop2"] == 24
        all_of_json_cp["obj2_prop1"] = 24
        assert json_results["metadata"]["allOf_test"]["obj3_switch"] == "obj3"
        assert json_results["metadata"]["allOf_test"]["obj3_prop1"] == 42.7
        assert json_results["metadata"]["allOf_test"]["obj3_prop2"] is True
        all_of_json_cp["obj3_prop1"] = 7.24

        with open(test_file, mode="w+") as all_of_fd:
            json.dump(all_of_json_cp, all_of_fd)

    # Replace the previously created instance
    ret = script_runner.run(
        "elyra-metadata",
        "update",
        METADATA_TEST_SCHEMASPACE,
        "--schema_name=metadata-test",
        f"--name={name}",
        f"--display_name=Test Complex {complex_keyword}2",
        "--required_test=required_value",
        f"{option}={value}",
    )
    assert ret.success
    assert f"Metadata instance '{name}' for schema 'metadata-test' has been written" in ret.stdout
    assert expected_file in ret.stdout
    assert os.path.exists(expected_file)

    with open(expected_file) as fd:
        json_results = json.load(fd)

    # Verify common stuff
    assert json_results["display_name"] == f"Test Complex {complex_keyword}2"
    assert json_results["metadata"]["required_test"] == "required_value"

    # Verify result following replace...
    if complex_keyword == "defs":
        assert json_results["metadata"]["defs_test"] == 24
    elif complex_keyword == "oneOf":
        assert json_results["metadata"]["oneOf_test"]["obj_switch"] == "obj2"
        assert json_results["metadata"]["oneOf_test"]["obj2_prop1"] == 24
        assert json_results["metadata"]["oneOf_test"]["obj2_prop2"] == 24
    elif complex_keyword == "allOf":
        assert len(json_results["metadata"]["allOf_test"]) == 9
        assert json_results["metadata"]["allOf_test"]["obj1_prop1"] == "allOf-test-val1-replace"
        assert json_results["metadata"]["allOf_test"]["obj1_prop2"] == "allOf-test-val2"
        assert json_results["metadata"]["allOf_test"]["obj2_prop1"] == 24
        assert json_results["metadata"]["allOf_test"]["obj2_prop2"] == 24
        assert json_results["metadata"]["allOf_test"]["obj3_prop1"] == 7.24
        assert json_results["metadata"]["allOf_test"]["obj3_prop2"] is True


# ---------- end of 'update' command tests


def test_list_help(script_runner):
    ret = script_runner.run("elyra-metadata", "list", METADATA_TEST_SCHEMASPACE, "--help")
    assert ret.success is False
    assert f"List installed metadata for {METADATA_TEST_SCHEMASPACE}." in ret.stdout


def test_list_bad_argument(script_runner):
    ret = script_runner.run("elyra-metadata", "list", METADATA_TEST_SCHEMASPACE, "--bogus-argument")
    assert ret.success is False
    assert "ERROR: The following arguments were unexpected: ['--bogus-argument']" in ret.stdout


def test_list_instances(script_runner, mock_data_dir):
    metadata_manager = MetadataManager(schemaspace=METADATA_TEST_SCHEMASPACE)

    ret = script_runner.run("elyra-metadata", "list", METADATA_TEST_SCHEMASPACE)
    assert ret.success
    lines = ret.stdout.split("\n")
    assert len(lines) == 2  # always 2 more than the actual runtime count
    assert f"No metadata instances found for {METADATA_TEST_SCHEMASPACE}" in lines[0]

    valid = Metadata(**valid_metadata_json)
    resource = metadata_manager.create("valid", valid)
    assert resource is not None
    resource = metadata_manager.create("valid2", valid)
    assert resource is not None
    another = Metadata(**another_metadata_json)
    resource = metadata_manager.create("another", another)
    assert resource is not None
    resource = metadata_manager.create("another2", another)
    assert resource is not None

    ret = script_runner.run("elyra-metadata", "list", METADATA_TEST_SCHEMASPACE)
    assert ret.success
    lines = ret.stdout.split("\n")
    assert len(lines) == 9  # always 5 more than the actual runtime count
    assert lines[0] == f"Available metadata instances for {METADATA_TEST_SCHEMASPACE} (includes invalid):"
    line_elements = [line.split() for line in lines[4:8]]
    assert line_elements[0][0] == "metadata-test"
    assert line_elements[0][1] == "another"
    assert line_elements[1][0] == "metadata-test"
    assert line_elements[1][1] == "another2"
    assert line_elements[2][0] == "metadata-test"
    assert line_elements[2][1] == "valid"
    assert line_elements[3][0] == "metadata-test"
    assert line_elements[3][1] == "valid2"

    # Remove the '2' runtimes and reconfirm smaller set
    metadata_manager.remove("valid2")
    metadata_manager.remove("another2")
    # Include two additional invalid files as well - one for uri failure, andother missing display_name
    metadata_dir = os.path.join(mock_data_dir, "metadata", METADATA_TEST_SCHEMASPACE)
    create_json_file(metadata_dir, "invalid.json", invalid_metadata_json)
    create_json_file(metadata_dir, "no_display_name.json", invalid_no_display_name_json)
    create_json_file(metadata_dir, "invalid_schema_name.json", invalid_schema_name_json)

    ret = script_runner.run("elyra-metadata", "list", METADATA_TEST_SCHEMASPACE)
    assert ret.success
    lines = ret.stdout.split("\n")
    assert len(lines) == 10  # always 5 more than the actual runtime count
    assert lines[0] == f"Available metadata instances for {METADATA_TEST_SCHEMASPACE} (includes invalid):"
    line_elements = [line.split() for line in lines[4:9]]
    assert line_elements[0][1] == "another"
    assert line_elements[1][1] == "invalid"
    assert line_elements[1][3] == "**INVALID**"
    assert line_elements[1][4] == "(ValidationError)"
    assert line_elements[2][3] == "**INVALID**"
    assert line_elements[2][4] == "(ValidationError)"
    assert line_elements[3][1] == "valid"
    assert line_elements[4][3] == "**INVALID**"
    assert line_elements[4][4] == "(SchemaNotFoundError)"

    ret = script_runner.run("elyra-metadata", "list", METADATA_TEST_SCHEMASPACE, "--valid-only")
    assert ret.success
    lines = ret.stdout.split("\n")
    assert len(lines) == 7  # always 5 more than the actual runtime count
    assert lines[0] == f"Available metadata instances for {METADATA_TEST_SCHEMASPACE} (valid only):"
    line_elements = [line.split() for line in lines[4:6]]
    assert line_elements[0][1] == "another"
    assert line_elements[1][1] == "valid"


def test_list_json_instances(script_runner, mock_data_dir):
    metadata_manager = MetadataManager(schemaspace=METADATA_TEST_SCHEMASPACE)

    ret = script_runner.run("elyra-metadata", "list", METADATA_TEST_SCHEMASPACE, "--json")
    assert ret.success
    lines = ret.stdout.split("\n")
    assert len(lines) == 2
    assert lines[0] == "[]"

    valid = Metadata(**valid_metadata_json)
    resource = metadata_manager.create("valid", valid)
    assert resource is not None
    resource = metadata_manager.create("valid2", valid)
    assert resource is not None
    another = Metadata(**another_metadata_json)
    resource = metadata_manager.create("another", another)
    assert resource is not None
    resource = metadata_manager.create("another2", another)
    assert resource is not None

    ret = script_runner.run("elyra-metadata", "list", METADATA_TEST_SCHEMASPACE, "--json")
    assert ret.success
    # Consume results
    results = json.loads(ret.stdout)
    assert len(results) == 4

    # Remove the '2' runtimes and reconfirm smaller set
    metadata_manager.remove("valid2")
    metadata_manager.remove("another2")

    # Include two additional invalid files as well - one for uri failure, andother missing display_name
    metadata_dir = os.path.join(mock_data_dir, "metadata", METADATA_TEST_SCHEMASPACE)
    create_json_file(metadata_dir, "invalid.json", invalid_metadata_json)
    create_json_file(metadata_dir, "no_display_name.json", invalid_no_display_name_json)

    ret = script_runner.run("elyra-metadata", "list", METADATA_TEST_SCHEMASPACE, "--json")
    assert ret.success
    results = json.loads(ret.stdout)
    assert len(results) == 4

    ret = script_runner.run("elyra-metadata", "list", METADATA_TEST_SCHEMASPACE, "--json", "--valid-only")
    assert ret.success
    results = json.loads(ret.stdout)
    assert len(results) == 2


def test_remove_help(script_runner):
    ret = script_runner.run("elyra-metadata", "remove", METADATA_TEST_SCHEMASPACE, "--help")
    assert ret.success is False
    assert f"Remove a metadata instance from schemaspace '{METADATA_TEST_SCHEMASPACE}'." in ret.stdout


def test_remove_no_name(script_runner):
    ret = script_runner.run("elyra-metadata", "remove", METADATA_TEST_SCHEMASPACE)
    assert ret.success is False
    assert "ERROR: '--name' is a required parameter." in ret.stdout


def test_remove_with_no_equals(script_runner):
    # Attempt removal w/o the '=' between parameter and value
    metadata_manager = MetadataManager(schemaspace=METADATA_TEST_SCHEMASPACE)
    valid = Metadata(**valid_metadata_json)
    metadata_manager.create("valid", valid)

    ret = script_runner.run("elyra-metadata", "remove", METADATA_TEST_SCHEMASPACE, "--name", "valid")
    assert ret.success is True


def test_remove_missing(script_runner, mock_data_dir):
    # Create an instance so that the schemaspace exists.
    metadata_manager = MetadataManager(schemaspace=METADATA_TEST_SCHEMASPACE)
    valid = Metadata(**valid_metadata_json)
    metadata_manager.create("valid", valid)

    ret = script_runner.run("elyra-metadata", "remove", METADATA_TEST_SCHEMASPACE, "--name=missing")
    assert ret.success is False
    assert "No such instance named 'missing' was found in the metadata-tests schemaspace." in ret.stdout

    # Now cleanup original instance.
    ret = script_runner.run("elyra-metadata", "remove", METADATA_TEST_SCHEMASPACE, "--name=valid")
    assert ret.success


def test_remove_instance(script_runner, mock_data_dir):
    metadata_manager = MetadataManager(schemaspace=METADATA_TEST_SCHEMASPACE)

    valid = Metadata(**valid_metadata_json)
    resource = metadata_manager.create("valid", valid)
    assert resource is not None
    resource = metadata_manager.create("valid2", valid)
    assert resource is not None
    another = Metadata(**another_metadata_json)
    resource = metadata_manager.create("another", another)
    assert resource is not None
    resource = metadata_manager.create("another2", another)
    assert resource is not None

    ret = script_runner.run("elyra-metadata", "remove", METADATA_TEST_SCHEMASPACE, "--name=valid")
    assert ret.success

    ret = script_runner.run("elyra-metadata", "remove", METADATA_TEST_SCHEMASPACE, "--name=another")
    assert ret.success

    instances = metadata_manager.get_all()
    assert len(instances) == 2
    assert instances[0].name.endswith("2")
    assert instances[1].name.endswith("2")


def test_export_help(script_runner):
    ret = script_runner.run("elyra-metadata", "export", METADATA_TEST_SCHEMASPACE, "--help")
    assert ret.success is False
    assert f"Export installed metadata in schemaspace '{METADATA_TEST_SCHEMASPACE}'" in ret.stdout


def test_export_no_directory(script_runner):
    ret = script_runner.run("elyra-metadata", "export", METADATA_TEST_SCHEMASPACE)
    assert ret.success is False
    assert "'--directory' is a required parameter." in ret.stdout


def test_export_bad_argument(script_runner):
    ret = script_runner.run(
        "elyra-metadata", "export", METADATA_TEST_SCHEMASPACE, "--directory=dummy-directory", "--bogus-argument"
    )
    assert ret.success is False
    assert "The following arguments were unexpected: ['--bogus-argument']" in ret.stdout


def test_export_bad_schemaspace(script_runner):
    ret = script_runner.run("elyra-metadata", "export", "bogus-schemaspace")
    assert ret.success is False
    assert "Subcommand 'bogus-schemaspace' is invalid." in ret.stdout


def test_export_bad_schema(script_runner):
    ret = script_runner.run(
        "elyra-metadata",
        "export",
        METADATA_TEST_SCHEMASPACE,
        "--directory=dummy-directory",
        "--schema_name=bogus-schema",
    )
    assert (
        "Schema name 'bogus-schema' is invalid. For the 'metadata-tests' schemaspace, "
        "the schema name must be one of ['metadata-test', 'metadata-test2']" in ret.stdout
    )
    assert ret.success is False


def test_export_no_schema_no_instances(script_runner, mock_data_dir):
    ret = script_runner.run("elyra-metadata", "export", METADATA_TEST_SCHEMASPACE, "--directory=dummy-directory")
    assert ret.success
    assert f"No metadata instances found for schemaspace '{METADATA_TEST_SCHEMASPACE}'" in ret.stdout
    assert "Nothing exported to 'dummy-directory'" in ret.stdout


def test_export_inaccessible_directory(script_runner, mock_data_dir):
    metadata_manager = MetadataManager(schemaspace=METADATA_TEST_SCHEMASPACE)

    # create metadata
    valid = Metadata(**valid_metadata_json)
    resource = metadata_manager.create("valid", valid)
    assert resource is not None

    directory_parameter = "/dummy-directory"

    ret = script_runner.run("elyra-metadata", "export", METADATA_TEST_SCHEMASPACE, f"--directory={directory_parameter}")
    assert ret.success is False
    assert f"Error creating directory structure for '{directory_parameter}/{METADATA_TEST_SCHEMASPACE}': " in ret.stdout
    assert any(ele in ret.stdout for ele in ["Read-only file system: ", "Permission denied: ", "Access Denied: "])
    assert f"'{directory_parameter}'" in ret.stdout


def test_export_with_schema_no_instances(script_runner, mock_data_dir):
    metadata_manager = MetadataManager(schemaspace=METADATA_TEST_SCHEMASPACE)

    # create metadata in a different schema
    valid = Metadata(**valid_metadata_json)
    resource = metadata_manager.create("valid", valid)
    assert resource is not None

    ret = script_runner.run(
        "elyra-metadata",
        "export",
        METADATA_TEST_SCHEMASPACE,
        "--schema_name=metadata-test2",
        "--directory=dummy-directory",
    )
    assert ret.success
    assert (
        f"No metadata instances found for schemaspace '{METADATA_TEST_SCHEMASPACE}' "
        f"and schema 'metadata-test2'" in ret.stdout
    )
    assert "Nothing exported to 'dummy-directory'" in ret.stdout


def test_export_no_schema_with_instances(script_runner, mock_data_dir):
    metadata_manager = MetadataManager(schemaspace=METADATA_TEST_SCHEMASPACE)

    # create valid metadata
    valid = Metadata(**valid_metadata_json)
    resource = metadata_manager.create("valid", valid)
    assert resource is not None

    # create invalid metadata
    metadata_dir = os.path.join(mock_data_dir, "metadata", METADATA_TEST_SCHEMASPACE)
    create_json_file(metadata_dir, "invalid.json", invalid_metadata_json)
    create_json_file(metadata_dir, "invalid2.json", invalid_metadata_json)

    # test for valid and invalid
    temp_dir = TemporaryDirectory()
    directory_parameter = temp_dir.name
    ret = script_runner.run(
        "elyra-metadata", "export", METADATA_TEST_SCHEMASPACE, "--include-invalid", f"--directory={directory_parameter}"
    )
    assert ret.success
    export_directory = os.path.join(directory_parameter, METADATA_TEST_SCHEMASPACE)
    assert f"Creating directory structure for '{export_directory}'" in ret.stdout
    assert (
        f"Exporting metadata instances for schemaspace '{METADATA_TEST_SCHEMASPACE}' "
        f"(includes invalid) to '{export_directory}'" in ret.stdout
    )
    assert "Exported 3 instances (2 of which are invalid)" in ret.stdout

    exported_metadata = sorted(os.listdir(export_directory), key=str.casefold)
    assert len(exported_metadata) == 3
    assert exported_metadata[0] == "invalid.json"
    assert exported_metadata[1] == "invalid2.json"
    assert exported_metadata[2] == "valid.json"
    temp_dir.cleanup()

    # test for valid and invalid using '--include-invalid' option, which
    # prior to version 4.0 is a no-op
    temp_dir = TemporaryDirectory()
    directory_parameter = temp_dir.name
    ret = script_runner.run(
        "elyra-metadata", "export", METADATA_TEST_SCHEMASPACE, f"--directory={directory_parameter}", "--include-invalid"
    )
    assert ret.success
    export_directory = os.path.join(directory_parameter, METADATA_TEST_SCHEMASPACE)
    assert f"Creating directory structure for '{export_directory}'" in ret.stdout
    assert (
        f"Exporting metadata instances for schemaspace '{METADATA_TEST_SCHEMASPACE}' "
        f"(includes invalid) to '{export_directory}'" in ret.stdout
    )
    assert "Exported 3 instances (2 of which are invalid)" in ret.stdout

    exported_metadata = sorted(os.listdir(export_directory), key=str.casefold)
    assert len(exported_metadata) == 3
    assert exported_metadata[0] == "invalid.json"
    assert exported_metadata[1] == "invalid2.json"
    assert exported_metadata[2] == "valid.json"
    temp_dir.cleanup()

    # test for valid only
    temp_dir = TemporaryDirectory()
    directory_parameter = temp_dir.name
    ret = script_runner.run("elyra-metadata", "export", METADATA_TEST_SCHEMASPACE, f"--directory={directory_parameter}")
    assert ret.success
    export_directory = os.path.join(directory_parameter, METADATA_TEST_SCHEMASPACE)
    assert f"Creating directory structure for '{export_directory}'" in ret.stdout
    assert (
        f"Exporting metadata instances for schemaspace '{METADATA_TEST_SCHEMASPACE}' "
        f"(valid only) to '{export_directory}'" in ret.stdout
    )
    assert "Exported 1 instance (0 of which are invalid)" in ret.stdout

    exported_metadata = os.listdir(export_directory)
    assert len(exported_metadata) == 1
    assert exported_metadata[0] == "valid.json"
    temp_dir.cleanup()


def test_export_with_schema_with_instances(script_runner, mock_data_dir):
    metadata_manager = MetadataManager(schemaspace=METADATA_TEST_SCHEMASPACE)

    # create valid metadata
    valid = Metadata(**valid_metadata_json)
    resource = metadata_manager.create("valid", valid)
    assert resource is not None
    valid = Metadata(**valid_metadata2_json)
    resource = metadata_manager.create("valid2", valid)
    assert resource is not None

    # create invalid metadata
    metadata_dir = os.path.join(mock_data_dir, "metadata", METADATA_TEST_SCHEMASPACE)
    create_json_file(metadata_dir, "invalid.json", invalid_metadata_json)

    # create export directory
    temp_dir = TemporaryDirectory()
    directory_parameter = temp_dir.name

    # test for valid and invalid
    ret = script_runner.run(
        "elyra-metadata",
        "export",
        METADATA_TEST_SCHEMASPACE,
        "--include-invalid",
        "--schema_name=metadata-test",
        f"--directory={directory_parameter}",
    )
    assert ret.success
    export_directory = os.path.join(directory_parameter, METADATA_TEST_SCHEMASPACE)
    assert f"Creating directory structure for '{export_directory}'" in ret.stdout
    assert (
        f"Exporting metadata instances for schemaspace '{METADATA_TEST_SCHEMASPACE}'"
        f" and schema 'metadata-test' (includes invalid) to '{export_directory}'" in ret.stdout
    )
    assert "Exported 2 instances (1 of which is invalid)" in ret.stdout

    exported_metadata = sorted(os.listdir(export_directory), key=str.casefold)
    assert len(exported_metadata) == 2
    assert exported_metadata[0] == "invalid.json"
    assert exported_metadata[1] == "valid.json"
    temp_dir.cleanup()

    # create export directory
    temp_dir = TemporaryDirectory()
    directory_parameter = temp_dir.name

    # test for valid only
    ret = script_runner.run(
        "elyra-metadata",
        "export",
        METADATA_TEST_SCHEMASPACE,
        "--schema_name=metadata-test",
        f"--directory={directory_parameter}",
    )
    assert ret.success
    export_directory = os.path.join(directory_parameter, METADATA_TEST_SCHEMASPACE)
    assert f"Creating directory structure for '{export_directory}'" in ret.stdout
    assert (
        f"Exporting metadata instances for schemaspace '{METADATA_TEST_SCHEMASPACE}'"
        f" and schema 'metadata-test' (valid only) to '{export_directory}'" in ret.stdout
    )
    assert "Exported 1 instance (0 of which are invalid)" in ret.stdout

    exported_metadata = os.listdir(export_directory)
    assert len(exported_metadata) == 1
    assert exported_metadata[0] == "valid.json"
    temp_dir.cleanup()


def test_export_without_clean(script_runner, mock_data_dir):
    metadata_manager = MetadataManager(schemaspace=METADATA_TEST_SCHEMASPACE)

    # create valid metadata
    valid = Metadata(**valid_metadata_json)
    resource = metadata_manager.create("valid", valid)
    assert resource is not None

    # create export directory
    temp_dir = TemporaryDirectory()
    directory_parameter = temp_dir.name
    export_directory = os.path.join(directory_parameter, METADATA_TEST_SCHEMASPACE)
    os.mkdir(export_directory)

    # create metadata file with dummy data and verify its contents
    dummy_json = {"required_test": "required_value"}
    metadata_filename = "valid.json"
    create_json_file(export_directory, metadata_filename, dummy_json)
    metadata_file_path = os.path.join(export_directory, metadata_filename)
    assert os.path.exists(metadata_file_path)
    assert json.loads(open(metadata_file_path).read()) == dummy_json

    # create additional dummy file with a different name and verify its contents
    dummy_filename = "dummy.json"
    create_json_file(export_directory, dummy_filename, dummy_json)
    dummy_file_path = os.path.join(export_directory, dummy_filename)
    assert os.path.exists(dummy_file_path)
    assert json.loads(open(dummy_file_path).read()) == dummy_json

    # create dummy file under different folder (different schema) and verify its contents
    export_directory_other = os.path.join(directory_parameter, "runtimes")
    os.mkdir(export_directory_other)
    dummy_filename_other = "dummy.json"
    create_json_file(export_directory_other, dummy_filename_other, dummy_json)
    dummy_file_path_other = os.path.join(export_directory_other, dummy_filename_other)
    assert os.path.exists(dummy_file_path_other)
    assert json.loads(open(dummy_file_path_other).read()) == dummy_json

    # export metadata without --clean flag
    ret = script_runner.run(
        "elyra-metadata",
        "export",
        METADATA_TEST_SCHEMASPACE,
        "--schema_name=metadata-test",
        f"--directory={directory_parameter}",
    )
    assert ret.success
    assert f"Creating directory structure for '{export_directory}'" not in ret.stdout
    assert (
        f"Exporting metadata instances for schemaspace '{METADATA_TEST_SCHEMASPACE}'"
        f" and schema 'metadata-test' (valid only) to '{export_directory}'" in ret.stdout
    )
    assert "Exported 1 instance (0 of which are invalid)" in ret.stdout

    # verify that the metadata file was overwritten while both the dummy files were left as is
    export_directory_files = sorted(os.listdir(export_directory), key=str.casefold)
    assert len(export_directory_files) == 2

    assert export_directory_files[0] == dummy_filename
    assert json.loads(open(dummy_file_path).read()) == dummy_json

    assert export_directory_files[1] == metadata_filename
    exported_metadata = json.loads(open(metadata_file_path).read())
    assert "schema_name" in exported_metadata
    assert exported_metadata.get("schema_name") == valid_metadata_json.get("schema_name")

    export_directory_other_files = sorted(os.listdir(export_directory_other), key=str.casefold)
    assert len(export_directory_other_files) == 1
    assert export_directory_other_files[0] == dummy_filename_other
    assert json.loads(open(dummy_file_path_other).read()) == dummy_json
    temp_dir.cleanup()


def test_export_clean(script_runner, mock_data_dir):
    metadata_manager = MetadataManager(schemaspace=METADATA_TEST_SCHEMASPACE)

    # create valid metadata
    valid = Metadata(**valid_metadata_json)
    resource = metadata_manager.create("valid", valid)
    assert resource is not None

    # create export directory
    temp_dir = TemporaryDirectory()
    directory_parameter = temp_dir.name
    export_directory = os.path.join(directory_parameter, METADATA_TEST_SCHEMASPACE)
    os.mkdir(export_directory)

    # create metadata file with dummy data and verify its contents
    dummy_json = {"required_test": "required_value"}
    metadata_filename = "valid.json"
    create_json_file(export_directory, metadata_filename, dummy_json)
    metadata_file_path = os.path.join(export_directory, metadata_filename)
    assert os.path.exists(metadata_file_path)
    assert json.loads(open(metadata_file_path).read()) == dummy_json

    # create additional dummy file with a different name and verify its contents
    dummy_filename = "dummy.json"
    create_json_file(export_directory, dummy_filename, dummy_json)
    dummy_file_path = os.path.join(export_directory, dummy_filename)
    assert os.path.exists(dummy_file_path)
    assert json.loads(open(dummy_file_path).read()) == dummy_json

    # create dummy file under different folder (different schema) and verify its contents
    export_directory_other = os.path.join(directory_parameter, "runtimes")
    os.mkdir(export_directory_other)
    dummy_filename_other = "dummy.json"
    create_json_file(export_directory_other, dummy_filename_other, dummy_json)
    dummy_file_path_other = os.path.join(export_directory_other, dummy_filename_other)
    assert os.path.exists(dummy_file_path_other)
    assert json.loads(open(dummy_file_path_other).read()) == dummy_json

    # export metadata with --clean flag
    ret = script_runner.run(
        "elyra-metadata",
        "export",
        METADATA_TEST_SCHEMASPACE,
        "--clean",
        "--schema_name=metadata-test",
        f"--directory={directory_parameter}",
    )
    assert ret.success
    assert f"Creating directory structure for '{export_directory}'" not in ret.stdout
    assert f"Cleaning out all files in '{export_directory}'" in ret.stdout
    assert (
        f"Exporting metadata instances for schemaspace '{METADATA_TEST_SCHEMASPACE}'"
        f" and schema 'metadata-test' (valid only) to '{export_directory}'" in ret.stdout
    )
    assert "Exported 1 instance (0 of which are invalid)" in ret.stdout

    # verify that the metadata file was overwritten and dummy file within the same schema folder was deleted
    # whereas the dummy file within the other schema folder was left as is
    export_directory_files = os.listdir(export_directory)
    assert len(export_directory_files) == 1

    assert export_directory_files[0] == metadata_filename
    exported_metadata = json.loads(open(metadata_file_path).read())
    assert "schema_name" in exported_metadata
    assert exported_metadata.get("schema_name") == valid_metadata_json.get("schema_name")

    export_directory_other_files = sorted(os.listdir(export_directory_other), key=str.casefold)
    assert len(export_directory_other_files) == 1
    assert export_directory_other_files[0] == dummy_filename_other
    assert json.loads(open(dummy_file_path_other).read()) == dummy_json
    temp_dir.cleanup()


def test_import_help(script_runner):
    ret = script_runner.run("elyra-metadata", "import", METADATA_TEST_SCHEMASPACE, "--help")
    assert ret.success is False
    assert f"\nImport metadata instances into schemaspace '{METADATA_TEST_SCHEMASPACE}'" in ret.stdout


def test_import_no_directory(script_runner):
    ret = script_runner.run("elyra-metadata", "import", METADATA_TEST_SCHEMASPACE)
    assert ret.success is False
    assert "ERROR: '--directory' is a required parameter." in ret.stdout


def test_import_bad_argument(script_runner):
    ret = script_runner.run(
        "elyra-metadata", "import", METADATA_TEST_SCHEMASPACE, "--directory=dummy-directory", "--bogus-argument"
    )
    assert ret.success is False
    assert "ERROR: The following arguments were unexpected: ['--bogus-argument']" in ret.stdout


def test_import_bad_schemaspace(script_runner):
    ret = script_runner.run("elyra-metadata", "import", "bogus-schemaspace")
    assert ret.success is False
    assert "Subcommand 'bogus-schemaspace' is invalid." in ret.stdout


def test_import_inaccessible_directory(script_runner):
    directory_parameter = "/dummy-directory"

    ret = script_runner.run("elyra-metadata", "import", METADATA_TEST_SCHEMASPACE, f"--directory={directory_parameter}")
    assert ret.success is False
    assert (
        f"Unable to reach the '{directory_parameter}'"
        f" directory: No such file or directory: '{directory_parameter}" in ret.stdout
    )


def test_import_empty_directory(script_runner):
    # create import directory
    temp_dir = TemporaryDirectory()
    directory_parameter = temp_dir.name

    # import metadata
    ret = script_runner.run("elyra-metadata", "import", METADATA_TEST_SCHEMASPACE, f"--directory={directory_parameter}")
    assert ret.success is True
    assert f"No instances for import found in the '{directory_parameter}' directory" in ret.stdout
    temp_dir.cleanup()


def test_import_non_json_file(script_runner):
    # create import directory
    temp_dir = TemporaryDirectory()
    directory_parameter = temp_dir.name

    # add a dummy file in the directory
    dummy_filename = "dummy.txt"
    dummy_filepath = os.path.join(directory_parameter, dummy_filename)
    dummy_file_content = "This is a dummy txt file."
    with open(dummy_filepath, "w") as f:
        f.write(dummy_file_content)
    assert os.path.exists(dummy_filepath)
    assert open(dummy_filepath).read() == dummy_file_content

    # import metadata
    ret = script_runner.run("elyra-metadata", "import", METADATA_TEST_SCHEMASPACE, f"--directory={directory_parameter}")
    assert ret.success is True
    assert f"No instances for import found in the '{directory_parameter}' directory" in ret.stdout
    temp_dir.cleanup()


def test_import_valid_metadata_files(script_runner, mock_data_dir):
    # create import directory
    temp_dir = TemporaryDirectory()
    directory_parameter = temp_dir.name

    # add valid metadata json files in the directory
    metadata_filename = "valid.json"
    metadata_file_path = os.path.join(directory_parameter, metadata_filename)
    with open(metadata_file_path, "w") as f:
        json.dump(valid_metadata_json, f)
    assert os.path.exists(metadata_file_path)
    assert json.loads(open(metadata_file_path).read()) == valid_metadata_json

    metadata_filename2 = "valid2.json"
    metadata_file_path2 = os.path.join(directory_parameter, metadata_filename2)
    with open(metadata_file_path2, "w") as f:
        json.dump(valid_metadata2_json, f)
    assert os.path.exists(metadata_file_path2)
    assert json.loads(open(metadata_file_path2).read()) == valid_metadata2_json

    # import metadata
    ret = script_runner.run("elyra-metadata", "import", METADATA_TEST_SCHEMASPACE, f"--directory={directory_parameter}")
    assert ret.success is True
    assert "Imported 2 instances" in ret.stdout
    temp_dir.cleanup()

    # verify contents of imported metadata
    metadata_directory = os.path.join(mock_data_dir, "metadata", METADATA_TEST_SCHEMASPACE)
    assert os.path.isdir(metadata_directory)
    installed_metadata_file_path = os.path.join(metadata_directory, metadata_filename)
    assert os.path.isfile(installed_metadata_file_path)

    with open(installed_metadata_file_path, "r") as metadata_file:
        instance_json = json.load(metadata_file)
        assert instance_json["schema_name"] == valid_metadata_json["schema_name"]
        assert instance_json["display_name"] == valid_metadata_json["display_name"]
        assert instance_json["metadata"]["required_test"] == valid_metadata_json["metadata"]["required_test"]
        assert instance_json["metadata"]["uri_test"] == valid_metadata_json["metadata"]["uri_test"]
        assert instance_json["metadata"]["number_range_test"] == valid_metadata_json["metadata"]["number_range_test"]

    installed_metadata_file_path = os.path.join(metadata_directory, metadata_filename2)
    assert os.path.isfile(installed_metadata_file_path)

    with open(installed_metadata_file_path, "r") as metadata_file:
        instance_json = json.load(metadata_file)
        assert instance_json["schema_name"] == valid_metadata2_json["schema_name"]
        assert instance_json["display_name"] == valid_metadata2_json["display_name"]
        assert instance_json["metadata"]["required_test"] == valid_metadata2_json["metadata"]["required_test"]
        assert instance_json["metadata"]["uri_test"] == valid_metadata2_json["metadata"]["uri_test"]
        assert instance_json["metadata"]["number_range_test"] == valid_metadata2_json["metadata"]["number_range_test"]


def test_import_invalid_metadata_file(script_runner, mock_data_dir):
    # create import directory
    temp_dir = TemporaryDirectory()
    directory_parameter = temp_dir.name

    # add invalid metadata json file in the directory
    metadata_filename = "invalid.json"
    metadata_file_path = os.path.join(directory_parameter, metadata_filename)
    with open(metadata_file_path, "w") as f:
        json.dump(invalid_metadata_json, f)
    assert os.path.exists(metadata_file_path)
    assert json.loads(open(metadata_file_path).read()) == invalid_metadata_json

    # import metadata
    ret = script_runner.run("elyra-metadata", "import", METADATA_TEST_SCHEMASPACE, f"--directory={directory_parameter}")
    assert ret.success is True
    lines = ret.stdout.split("\n")
    assert len(lines) == 8
    assert "Imported 0 instances" in lines[0]
    assert "1 instance could not be imported" in lines[1]
    assert "The following files could not be imported:" in lines[3]
    assert lines[6].startswith("invalid.json")
    assert (
        lines[6]
        .strip()
        .endswith(
            "Validation failed for instance 'invalid' using the metadata-test "
            + "schema with error: '//localhost:8081/' is not a 'uri'."
        )
    )

    temp_dir.cleanup()


def test_import_with_subfolder(script_runner, mock_data_dir):
    # create import directory
    temp_dir = TemporaryDirectory()
    directory_parameter = temp_dir.name

    # add valid metadata json file in the import directory
    metadata_filename = "valid.json"
    metadata_file_path = os.path.join(directory_parameter, metadata_filename)
    with open(metadata_file_path, "w") as f:
        json.dump(valid_metadata_json, f)
    assert os.path.exists(metadata_file_path)
    assert json.loads(open(metadata_file_path).read()) == valid_metadata_json

    # add invalid metadata json files in the directory
    invalid_metadata_filename = "invalid.json"
    invalid_metadata_file_path = os.path.join(directory_parameter, invalid_metadata_filename)
    with open(invalid_metadata_file_path, "w") as f:
        json.dump(invalid_metadata_json, f)
    assert os.path.exists(invalid_metadata_file_path)
    assert json.loads(open(invalid_metadata_file_path).read()) == invalid_metadata_json

    invalid_metadata_file_path2 = os.path.join(directory_parameter, "invalid2.json")
    shutil.copyfile(invalid_metadata_file_path, invalid_metadata_file_path2)
    assert os.path.exists(invalid_metadata_file_path2)
    assert json.loads(open(invalid_metadata_file_path2).read()) == invalid_metadata_json

    # create a sub-folder within import directory and add a valid metadata file in it
    os.mkdir(os.path.join(directory_parameter, "subfolder"))
    metadata_filename2 = "valid2.json"
    metadata_file_path2 = os.path.join(directory_parameter, "subfolder", metadata_filename2)
    with open(metadata_file_path2, "w") as f:
        json.dump(valid_metadata2_json, f)
    assert os.path.exists(metadata_file_path2)
    assert json.loads(open(metadata_file_path2).read()) == valid_metadata2_json

    # import metadata
    ret = script_runner.run("elyra-metadata", "import", METADATA_TEST_SCHEMASPACE, f"--directory={directory_parameter}")
    assert ret.success is True
    lines = ret.stdout.split("\n")
    assert len(lines) == 9
    assert "Imported 1 instance" in lines[0]
    assert "2 instances could not be imported" in lines[1]
    assert "The following files could not be imported:" in lines[3]
    assert lines[6].startswith("invalid.json")
    assert (
        lines[6]
        .strip()
        .endswith(
            "Validation failed for instance 'invalid' using the metadata-test "
            + "schema with error: '//localhost:8081/' is not a 'uri'."
        )
    )
    assert lines[7].startswith("invalid2.json")
    assert (
        lines[7]
        .strip()
        .endswith(
            "Validation failed for instance 'invalid2' using the metadata-test "
            + "schema with error: '//localhost:8081/' is not a 'uri'."
        )
    )

    temp_dir.cleanup()

    # verify contents of imported metadata
    assert os.path.isdir(os.path.join(mock_data_dir, "metadata", METADATA_TEST_SCHEMASPACE))
    installed_metadata_file_path = os.path.join(mock_data_dir, "metadata", METADATA_TEST_SCHEMASPACE, metadata_filename)
    assert os.path.isfile(installed_metadata_file_path)

    with open(installed_metadata_file_path, "r") as metadata_file:
        instance_json = json.load(metadata_file)
        assert instance_json["schema_name"] == valid_metadata_json["schema_name"]
        assert instance_json["display_name"] == valid_metadata_json["display_name"]
        assert instance_json["metadata"]["required_test"] == valid_metadata_json["metadata"]["required_test"]
        assert instance_json["metadata"]["uri_test"] == valid_metadata_json["metadata"]["uri_test"]
        assert instance_json["metadata"]["number_range_test"] == valid_metadata_json["metadata"]["number_range_test"]


def test_import_overwrite_flag(script_runner, mock_data_dir):
    metadata_manager = MetadataManager(schemaspace=METADATA_TEST_SCHEMASPACE)

    # create valid metadata
    valid = Metadata(**valid_metadata_json)
    resource = metadata_manager.create("valid", valid)
    assert resource is not None

    # create import directory
    temp_dir = TemporaryDirectory()
    directory_parameter = temp_dir.name

    # add valid metadata json file for existing metadata in the import directory
    metadata_filename = "valid.json"
    metadata_file_path = os.path.join(directory_parameter, metadata_filename)
    with open(metadata_file_path, "w") as f:
        json.dump(valid_metadata_json, f)
    assert os.path.exists(metadata_file_path)
    assert json.loads(open(metadata_file_path).read()) == valid_metadata_json

    # add valid metadata json file for new metadata
    metadata_filename2 = "valid2.json"
    metadata_file_path2 = os.path.join(directory_parameter, metadata_filename2)
    with open(metadata_file_path2, "w") as f:
        json.dump(valid_metadata_json, f)
    assert os.path.exists(metadata_file_path2)
    assert json.loads(open(metadata_file_path2).read()) == valid_metadata_json

    # import metadata without overwrite flag
    ret = script_runner.run("elyra-metadata", "import", METADATA_TEST_SCHEMASPACE, f"--directory={directory_parameter}")
    assert ret.success is True
    lines = ret.stdout.split("\n")
    assert len(lines) == 8
    assert "Imported 1 instance" in lines[0]
    assert "1 instance could not be imported" in lines[1]
    assert "The following files could not be imported:" in lines[3]
    assert lines[6].startswith("valid.json")
    assert (
        lines[6]
        .strip()
        .endswith(
            "An instance named 'valid' already exists in the metadata-tests "
            + "schemaspace. Use '--overwrite' to update."
        )
    )

    # verify contents of imported metadata
    assert os.path.isdir(os.path.join(mock_data_dir, "metadata", METADATA_TEST_SCHEMASPACE))
    installed_metadata_file_path2 = os.path.join(
        mock_data_dir, "metadata", METADATA_TEST_SCHEMASPACE, metadata_filename2
    )
    assert os.path.isfile(installed_metadata_file_path2)

    with open(installed_metadata_file_path2, "r") as metadata_file:
        instance_json = json.load(metadata_file)
        assert instance_json["schema_name"] == valid_metadata_json["schema_name"]
        assert instance_json["display_name"] == valid_metadata_json["display_name"]
        assert instance_json["metadata"]["required_test"] == valid_metadata_json["metadata"]["required_test"]
        assert instance_json["metadata"]["uri_test"] == valid_metadata_json["metadata"]["uri_test"]
        assert instance_json["metadata"]["number_range_test"] == valid_metadata_json["metadata"]["number_range_test"]

    # replace one of the metadata files with new content
    os.remove(metadata_file_path2)
    assert os.path.exists(metadata_file_path2) is False
    with open(metadata_file_path2, "w") as f:
        json.dump(valid_display_name_json, f)
    assert os.path.exists(metadata_file_path2)
    assert json.loads(open(metadata_file_path2).read()) == valid_display_name_json

    # add another valid metadata json file for new metadata
    metadata_filename3 = "another.json"
    metadata_file_path3 = os.path.join(directory_parameter, metadata_filename3)
    with open(metadata_file_path3, "w") as f:
        json.dump(another_metadata_json, f)
    assert os.path.exists(metadata_file_path3)
    assert json.loads(open(metadata_file_path3).read()) == another_metadata_json

    # re-try import metadata with overwrite flag
    ret = script_runner.run(
        "elyra-metadata",
        "import",
        METADATA_TEST_SCHEMASPACE,
        f"--directory={directory_parameter}",
        "--overwrite",
    )
    assert ret.success is True
    assert "Imported 3 instances" in ret.stdout
    temp_dir.cleanup()

    # verify contents of existing (unchanged) metadata
    assert os.path.isdir(os.path.join(mock_data_dir, "metadata", METADATA_TEST_SCHEMASPACE))
    installed_metadata_file_path = os.path.join(mock_data_dir, "metadata", METADATA_TEST_SCHEMASPACE, metadata_filename)
    assert os.path.isfile(installed_metadata_file_path)

    with open(installed_metadata_file_path, "r") as metadata_file:
        instance_json = json.load(metadata_file)
        assert instance_json["schema_name"] == valid_metadata_json["schema_name"]
        assert instance_json["display_name"] == valid_metadata_json["display_name"]
        assert instance_json["metadata"]["required_test"] == valid_metadata_json["metadata"]["required_test"]
        assert instance_json["metadata"]["uri_test"] == valid_metadata_json["metadata"]["uri_test"]
        assert instance_json["metadata"]["number_range_test"] == valid_metadata_json["metadata"]["number_range_test"]

    # verify contents of overwritten metadata
    installed_metadata_filepath2 = os.path.join(
        mock_data_dir, "metadata", METADATA_TEST_SCHEMASPACE, metadata_filename2
    )
    assert os.path.isfile(installed_metadata_filepath2)

    with open(installed_metadata_filepath2, "r") as metadata_file:
        instance_json = json.load(metadata_file)
        assert instance_json["schema_name"] == valid_display_name_json["schema_name"]
        assert instance_json["display_name"] == valid_display_name_json["display_name"]
        assert instance_json["metadata"]["required_test"] == valid_display_name_json["metadata"]["required_test"]

    # verify contents of new imported metadata
    installed_metadata_filepath3 = os.path.join(
        mock_data_dir, "metadata", METADATA_TEST_SCHEMASPACE, metadata_filename3
    )
    assert os.path.isfile(installed_metadata_filepath3)

    with open(installed_metadata_filepath3, "r") as metadata_file:
        instance_json = json.load(metadata_file)
        assert instance_json["schema_name"] == another_metadata_json["schema_name"]
        assert instance_json["display_name"] == another_metadata_json["display_name"]
        assert instance_json["metadata"]["required_test"] == another_metadata_json["metadata"]["required_test"]
        assert instance_json["metadata"]["uri_test"] == another_metadata_json["metadata"]["uri_test"]


# Begin property tests...


def test_required(script_runner, mock_data_dir):
    # Doesn't use PropertyTester due to its unique test since all other tests require this property
    name = "required"

    expected_file = os.path.join(mock_data_dir, "metadata", METADATA_TEST_SCHEMASPACE, name + ".json")
    # Cleanup from any potential previous failures
    if os.path.exists(expected_file):
        os.remove(expected_file)

    ret = script_runner.run(
        "elyra-metadata",
        "install",
        METADATA_TEST_SCHEMASPACE,
        "--schema_name=metadata-test",
        "--name=" + name,
        "--display_name=" + name,
    )

    assert ret.success is False
    assert "'--required_test' is a required parameter" in ret.stdout

    ret = script_runner.run(
        "elyra-metadata",
        "install",
        METADATA_TEST_SCHEMASPACE,
        "--schema_name=metadata-test",
        "--name=" + name,
        "--display_name=" + name,
        "--required_test=required_value",
    )

    assert ret.success
    assert "Metadata instance '" + name + "' for schema 'metadata-test' has been written" in ret.stdout

    assert os.path.isdir(os.path.join(mock_data_dir, "metadata", METADATA_TEST_SCHEMASPACE))
    assert os.path.isfile(expected_file)

    with open(expected_file, "r") as fd:
        instance_json = json.load(fd)
        assert instance_json["schema_name"] == "metadata-test"
        assert instance_json["display_name"] == name
        assert instance_json["metadata"]["required_test"] == "required_value"


def test_number_default(script_runner, mock_data_dir):
    # Doesn't use PropertyTester due to its unique test (no failure, needs --replace, etc.)
    name = "number_default"

    expected_file = os.path.join(mock_data_dir, "metadata", METADATA_TEST_SCHEMASPACE, name + ".json")
    # Cleanup from any potential previous failures
    if os.path.exists(expected_file):
        os.remove(expected_file)

    # No negative test here.  First create w/o a value and ensure 42, then create with a value and ensure that value.
    ret = script_runner.run(
        "elyra-metadata",
        "install",
        METADATA_TEST_SCHEMASPACE,
        "--schema_name=metadata-test",
        "--name=" + name,
        "--display_name=" + name,
        "--required_test=required_value",
    )

    assert ret.success
    assert "Metadata instance '" + name + "' for schema 'metadata-test' has been written" in ret.stdout

    assert os.path.isdir(os.path.join(mock_data_dir, "metadata", METADATA_TEST_SCHEMASPACE))
    assert os.path.isfile(expected_file)

    with open(expected_file, "r") as fd:
        instance_json = json.load(fd)
        assert instance_json["schema_name"] == "metadata-test"
        assert instance_json["display_name"] == name
        assert instance_json["metadata"]["number_default_test"] == 42

    # Note that we only include the properties that are changed, along with "identifiers" like name ans schema_name.
    ret = script_runner.run(
        "elyra-metadata",
        "install",
        METADATA_TEST_SCHEMASPACE,
        "--schema_name=metadata-test",
        "--name=" + name,
        "--replace",
        "--number_default_test=7.2",
    )

    assert ret.success
    assert "Metadata instance '" + name + "' for schema 'metadata-test' has been written" in ret.stdout

    assert os.path.isdir(os.path.join(mock_data_dir, "metadata", METADATA_TEST_SCHEMASPACE))
    assert os.path.isfile(expected_file)

    with open(expected_file, "r") as fd:
        instance_json = json.load(fd)
        assert instance_json["schema_name"] == "metadata-test"
        assert instance_json["display_name"] == name
        assert instance_json["metadata"]["number_default_test"] == 7.2


def test_uri(script_runner, mock_data_dir):
    prop_test = PropertyTester("uri")
    prop_test.negative_value = "//invalid-uri"
    prop_test.negative_stdout = "'//invalid-uri' is not a 'uri'"
    #  this can be joined with previous if adding meta-properties
    #  "; title: URI Test, format: uri"
    prop_test.negative_stderr = "'//invalid-uri' is not a 'uri'"
    prop_test.positive_value = "http://localhost:31823/v1/models?version=2017-02-13"
    prop_test.run(script_runner, mock_data_dir)


def test_integer_exclusivity(script_runner, mock_data_dir):
    prop_test = PropertyTester("integer_exclusivity")
    prop_test.negative_value = 3
    prop_test.negative_stdout = "3 is less than or equal to the minimum of 3"
    #  this can be joined with previous if adding meta-properties
    #  "; title: Integer Exclusivity Test, exclusiveMinimum: 3, exclusiveMaximum: 10"
    prop_test.negative_stderr = "3 is less than or equal to the minimum of 3"
    prop_test.positive_value = 7
    prop_test.run(script_runner, mock_data_dir)


def test_integer_multiple(script_runner, mock_data_dir):
    prop_test = PropertyTester("integer_multiple")
    prop_test.negative_value = 32
    prop_test.negative_stdout = "32 is not a multiple of 6"
    #  this can be joined with previous if adding meta-properties
    #  "; title: Integer Multiple Test, multipleOf: 6"
    prop_test.negative_stderr = "32 is not a multiple of 6"
    prop_test.positive_value = 42
    prop_test.run(script_runner, mock_data_dir)


def test_number_range(script_runner, mock_data_dir):
    prop_test = PropertyTester("number_range")
    prop_test.negative_value = 2.7
    prop_test.negative_stdout = "2.7 is less than the minimum of 3"
    #  this can be joined with previous if adding meta-properties
    #  "; title: Number Range Test, minimum: 3, maximum: 10"
    prop_test.negative_stderr = "2.7 is less than the minimum of 3"
    prop_test.positive_value = 7.2
    prop_test.run(script_runner, mock_data_dir)


def test_const(script_runner, mock_data_dir):
    prop_test = PropertyTester("const")
    prop_test.negative_value = 2.718
    prop_test.negative_stdout = "3.14 was expected"
    #  this can be joined with previous if adding meta-properties
    #  " ; title: Const Test, const: 3.14"
    prop_test.negative_stderr = "3.14 was expected"
    prop_test.positive_value = 3.14
    prop_test.run(script_runner, mock_data_dir)


def test_string_length(script_runner, mock_data_dir):
    prop_test = PropertyTester("string_length")
    prop_test.negative_value = "12345678901"
    prop_test.negative_stdout = "'12345678901' is too long"
    #  this can be joined with previous if adding meta-properties
    #  "; title: String Length Test, minLength: 3, maxLength: 10"
    prop_test.negative_stderr = "'12345678901' is too long"
    prop_test.positive_value = "123456"
    prop_test.run(script_runner, mock_data_dir)


def test_string_pattern(script_runner, mock_data_dir):
    prop_test = PropertyTester("string_pattern")  # Must start/end with alphanumeric, can include '-' and '.'
    prop_test.negative_value = "-foo1"
    prop_test.negative_stdout = "'-foo1' does not match '^[a-z0-9][a-z0-9-.]*[a-z0-9]$'"
    #  this can be joined with previous if adding meta-properties
    #  "; title: String Pattern Test, pattern: ^[a-z0-9][a-z0-9-.]*[a-z0-9]$"
    prop_test.negative_stderr = "'-foo1' does not match '^[a-z0-9][a-z0-9-.]*[a-z0-9]$'"
    prop_test.positive_value = "0foo-bar.com-01"
    prop_test.run(script_runner, mock_data_dir)


def test_enum(script_runner, mock_data_dir):
    prop_test = PropertyTester("enum")
    prop_test.negative_value = "jupyter"
    prop_test.negative_stdout = "'jupyter' is not one of ['elyra', 'rocks', 'added']"
    #  this can be joined with previous if adding meta-properties
    #  "; title: Enum Test, enum: ['elyra', 'rocks', 'added']"
    prop_test.negative_stderr = "'jupyter' is not one of ['elyra', 'rocks', 'added']"
    prop_test.positive_value = "added"
    prop_test.run(script_runner, mock_data_dir)


def test_array(script_runner, mock_data_dir):
    prop_test = PropertyTester("array")
    prop_test.negative_value = [1, 2, 2]
    prop_test.negative_stdout = "[1, 2, 2] has non-unique elements"
    #  this can be joined with previous if adding meta-properties
    #  "; title: Array Test, minItems: 3, maxItems: 10, uniqueItems: True"
    prop_test.negative_stderr = "[1, 2, 2] has non-unique elements"
    prop_test.positive_value = [1, 2, 3, 4, 5]
    prop_test.run(script_runner, mock_data_dir)


def test_object(script_runner, mock_data_dir):
    prop_test = PropertyTester("object")
    prop_test.negative_value = {"prop1": 2, "prop2": 3}
    prop_test.negative_stdout = "{'prop1': 2, 'prop2': 3} does not have enough properties"
    #  this can be joined with previous if adding meta-properties
    #  "; title: Object Test, minProperties: 3, maxProperties: 10"
    prop_test.negative_stderr = "{'prop1': 2, 'prop2': 3} does not have enough properties"
    prop_test.positive_value = {"prop1": 2, "prop2": 3, "prop3": 4, "prop4": 5}
    prop_test.run(script_runner, mock_data_dir)


def test_boolean(script_runner, mock_data_dir):
    prop_test = PropertyTester("boolean")
    prop_test.negative_value = "bogus_boolean"
    prop_test.negative_stdout = "'bogus_boolean' is not of type 'boolean'"
    #  this can be joined with previous if adding meta-properties
    #  "; title: Boolean Test"
    prop_test.negative_stderr = "'bogus_boolean' is not of type 'boolean'"
    prop_test.positive_value = True
    prop_test.run(script_runner, mock_data_dir)


def test_null(script_runner, mock_data_dir):
    prop_test = PropertyTester("null")
    prop_test.negative_value = "bogus_null"
    prop_test.negative_stdout = "'bogus_null' is not of type 'null'"
    #  this can be joined with previous if adding meta-properties
    #  "; title: Null Test"
    prop_test.negative_stderr = "'bogus_null' is not of type 'null'"
    prop_test.positive_value = None
    prop_test.run(script_runner, mock_data_dir)



#
# Copyright 2018-2022 Elyra Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
import string

from kfp.dsl import RUN_ID_PLACEHOLDER
import pytest

from elyra.kfp.operator import ExecuteFileOp


def test_fail_without_cos_endpoint():
    with pytest.raises(TypeError):
        ExecuteFileOp(
            name="test",
            pipeline_name="test-pipeline",
            experiment_name="experiment-name",
            notebook="test_notebook.ipynb",
            cos_bucket="test_bucket",
            cos_directory="test_directory",
            cos_dependencies_archive="test_archive.tgz",
            image="test/image:dev",
        )


def test_fail_without_cos_bucket():
    with pytest.raises(TypeError):
        ExecuteFileOp(
            name="test",
            pipeline_name="test-pipeline",
            experiment_name="experiment-name",
            notebook="test_notebook.ipynb",
            cos_endpoint="http://testserver:32525",
            cos_directory="test_directory",
            cos_dependencies_archive="test_archive.tgz",
            image="test/image:dev",
        )


def test_fail_without_cos_directory():
    with pytest.raises(TypeError):
        ExecuteFileOp(
            name="test",
            pipeline_name="test-pipeline",
            experiment_name="experiment-name",
            notebook="test_notebook.ipynb",
            cos_endpoint="http://testserver:32525",
            cos_bucket="test_bucket",
            cos_dependencies_archive="test_archive.tgz",
            image="test/image:dev",
        )


def test_fail_without_cos_dependencies_archive():
    with pytest.raises(TypeError):
        ExecuteFileOp(
            name="test",
            pipeline_name="test-pipeline",
            experiment_name="experiment-name",
            notebook="test_notebook.ipynb",
            cos_endpoint="http://testserver:32525",
            cos_bucket="test_bucket",
            cos_directory="test_directory",
            image="test/image:dev",
        )


def test_fail_without_runtime_image():
    with pytest.raises(ValueError) as error_info:
        ExecuteFileOp(
            name="test",
            pipeline_name="test-pipeline",
            experiment_name="experiment-name",
            notebook="test_notebook.ipynb",
            cos_endpoint="http://testserver:32525",
            cos_bucket="test_bucket",
            cos_directory="test_directory",
            cos_dependencies_archive="test_archive.tgz",
        )
    assert "You need to provide an image." == str(error_info.value)


def test_fail_without_notebook():
    with pytest.raises(TypeError):
        ExecuteFileOp(
            name="test",
            pipeline_name="test-pipeline",
            experiment_name="experiment-name",
            cos_endpoint="http://testserver:32525",
            cos_bucket="test_bucket",
            cos_directory="test_directory",
            cos_dependencies_archive="test_archive.tgz",
            image="test/image:dev",
        )


def test_fail_without_name():
    with pytest.raises(TypeError):
        ExecuteFileOp(
            pipeline_name="test-pipeline",
            experiment_name="experiment-name",
            notebook="test_notebook.ipynb",
            cos_endpoint="http://testserver:32525",
            cos_bucket="test_bucket",
            cos_directory="test_directory",
            cos_dependencies_archive="test_archive.tgz",
            image="test/image:dev",
        )


def test_fail_with_empty_string_as_name():
    with pytest.raises(ValueError):
        ExecuteFileOp(
            name="",
            pipeline_name="test-pipeline",
            experiment_name="experiment-name",
            notebook="test_notebook.ipynb",
            cos_endpoint="http://testserver:32525",
            cos_bucket="test_bucket",
            cos_directory="test_directory",
            cos_dependencies_archive="test_archive.tgz",
            image="test/image:dev",
        )


def test_fail_with_empty_string_as_notebook():
    with pytest.raises(ValueError) as error_info:
        ExecuteFileOp(
            name="test",
            pipeline_name="test-pipeline",
            experiment_name="experiment-name",
            notebook="",
            cos_endpoint="http://testserver:32525",
            cos_bucket="test_bucket",
            cos_directory="test_directory",
            cos_dependencies_archive="test_archive.tgz",
            image="test/image:dev",
        )
    assert "You need to provide a notebook." == str(error_info.value)


def test_fail_without_pipeline_name():
    with pytest.raises(TypeError):
        ExecuteFileOp(
            name="test",
            experiment_name="experiment-name",
            notebook="test_notebook.ipynb",
            cos_endpoint="http://testserver:32525",
            cos_bucket="test_bucket",
            cos_directory="test_directory",
            cos_dependencies_archive="test_archive.tgz",
            image="test/image:dev",
        )


def test_fail_without_experiment_name():
    with pytest.raises(TypeError):
        ExecuteFileOp(
            name="test",
            pipeline_name="test-pipeline",
            notebook="test_notebook.ipynb",
            cos_endpoint="http://testserver:32525",
            cos_bucket="test_bucket",
            cos_directory="test_directory",
            cos_dependencies_archive="test_archive.tgz",
            image="test/image:dev",
        )


def test_properly_set_notebook_name_when_in_subdirectory():
    notebook_op = ExecuteFileOp(
        name="test",
        pipeline_name="test-pipeline",
        experiment_name="experiment-name",
        notebook="foo/test_notebook.ipynb",
        cos_endpoint="http://testserver:32525",
        cos_bucket="test_bucket",
        cos_directory="test_directory",
        cos_dependencies_archive="test_archive.tgz",
        image="test/image:dev",
    )
    assert "test_notebook.ipynb" == notebook_op.notebook_name


def test_properly_set_python_script_name_when_in_subdirectory():
    notebook_op = ExecuteFileOp(
        name="test",
        pipeline_name="test-pipeline",
        experiment_name="experiment-name",
        notebook="foo/test.py",
        cos_endpoint="http://testserver:32525",
        cos_bucket="test_bucket",
        cos_directory="test_directory",
        cos_dependencies_archive="test_archive.tgz",
        image="test/image:dev",
    )
    assert "test.py" == notebook_op.notebook_name


def test_user_crio_volume_creation():
    notebook_op = ExecuteFileOp(
        name="test",
        pipeline_name="test-pipeline",
        experiment_name="experiment-name",
        notebook="test_notebook.ipynb",
        cos_endpoint="http://testserver:32525",
        cos_bucket="test_bucket",
        cos_directory="test_directory",
        cos_dependencies_archive="test_archive.tgz",
        image="test/image:dev",
        emptydir_volume_size="20Gi",
    )
    assert notebook_op.emptydir_volume_size == "20Gi"
    assert notebook_op.container_work_dir_root_path == "/opt/app-root/src/"
    assert notebook_op.container.volume_mounts.__len__() == 1
    # Environment variables: PYTHONPATH, ELYRA_RUN_NAME
    assert notebook_op.container.env.__len__() == 2, notebook_op.container.env


def test_override_bootstrap_url():
    notebook_op = ExecuteFileOp(
        name="test",
        pipeline_name="test-pipeline",
        experiment_name="experiment-name",
        bootstrap_script_url="https://test.server.com/bootscript.py",
        notebook="test_notebook.ipynb",
        cos_endpoint="http://testserver:32525",
        cos_bucket="test_bucket",
        cos_directory="test_directory",
        cos_dependencies_archive="test_archive.tgz",
        image="test/image:dev",
    )
    assert notebook_op.bootstrap_script_url == "https://test.server.com/bootscript.py"


def test_override_requirements_url():
    notebook_op = ExecuteFileOp(
        name="test",
        pipeline_name="test-pipeline",
        experiment_name="experiment-name",
        requirements_url="https://test.server.com/requirements.py",
        notebook="test_notebook.ipynb",
        cos_endpoint="http://testserver:32525",
        cos_bucket="test_bucket",
        cos_directory="test_directory",
        cos_dependencies_archive="test_archive.tgz",
        image="test/image:dev",
    )
    assert notebook_op.requirements_url == "https://test.server.com/requirements.py"


def test_construct_with_both_pipeline_inputs_and_outputs():
    notebook_op = ExecuteFileOp(
        name="test",
        pipeline_name="test-pipeline",
        experiment_name="experiment-name",
        notebook="test_notebook.ipynb",
        cos_endpoint="http://testserver:32525",
        cos_bucket="test_bucket",
        cos_directory="test_directory",
        cos_dependencies_archive="test_archive.tgz",
        pipeline_inputs=["test_input1.txt", "test_input2.txt"],
        pipeline_outputs=["test_output1.txt", "test_output2.txt"],
        image="test/image:dev",
    )
    assert notebook_op.pipeline_inputs == ["test_input1.txt", "test_input2.txt"]
    assert notebook_op.pipeline_outputs == ["test_output1.txt", "test_output2.txt"]

    assert '--inputs "test_input1.txt;test_input2.txt"' in notebook_op.container.args[0]
    assert '--outputs "test_output1.txt;test_output2.txt"' in notebook_op.container.args[0]


def test_construct_wildcard_outputs():
    notebook_op = ExecuteFileOp(
        name="test",
        pipeline_name="test-pipeline",
        experiment_name="experiment-name",
        notebook="test_notebook.ipynb",
        cos_endpoint="http://testserver:32525",
        cos_bucket="test_bucket",
        cos_directory="test_directory",
        cos_dependencies_archive="test_archive.tgz",
        pipeline_inputs=["test_input1.txt", "test_input2.txt"],
        pipeline_outputs=["test_out*", "foo.tar"],
        image="test/image:dev",
    )
    assert notebook_op.pipeline_inputs == ["test_input1.txt", "test_input2.txt"]
    assert notebook_op.pipeline_outputs == ["test_out*", "foo.tar"]

    assert '--inputs "test_input1.txt;test_input2.txt"' in notebook_op.container.args[0]
    assert '--outputs "test_out*;foo.tar"' in notebook_op.container.args[0]


def test_construct_with_only_pipeline_inputs():
    notebook_op = ExecuteFileOp(
        name="test",
        pipeline_name="test-pipeline",
        experiment_name="experiment-name",
        notebook="test_notebook.ipynb",
        cos_endpoint="http://testserver:32525",
        cos_bucket="test_bucket",
        cos_directory="test_directory",
        cos_dependencies_archive="test_archive.tgz",
        pipeline_inputs=["test_input1.txt", "test,input2.txt"],
        pipeline_outputs=[],
        image="test/image:dev",
    )
    assert notebook_op.pipeline_inputs == ["test_input1.txt", "test,input2.txt"]
    assert '--inputs "test_input1.txt;test,input2.txt"' in notebook_op.container.args[0]


def test_construct_with_bad_pipeline_inputs():
    with pytest.raises(ValueError) as error_info:
        ExecuteFileOp(
            name="test",
            pipeline_name="test-pipeline",
            experiment_name="experiment-name",
            notebook="test_notebook.ipynb",
            cos_endpoint="http://testserver:32525",
            cos_bucket="test_bucket",
            cos_directory="test_directory",
            cos_dependencies_archive="test_archive.tgz",
            pipeline_inputs=["test_input1.txt", "test;input2.txt"],
            pipeline_outputs=[],
            image="test/image:dev",
        )
    assert "Illegal character (;) found in filename 'test;input2.txt'." == str(error_info.value)


def test_construct_with_only_pipeline_outputs():
    notebook_op = ExecuteFileOp(
        name="test",
        pipeline_name="test-pipeline",
        experiment_name="experiment-name",
        notebook="test_notebook.ipynb",
        cos_endpoint="http://testserver:32525",
        cos_bucket="test_bucket",
        cos_directory="test_directory",
        cos_dependencies_archive="test_archive.tgz",
        pipeline_outputs=["test_output1.txt", "test,output2.txt"],
        pipeline_envs={},
        image="test/image:dev",
    )
    assert notebook_op.pipeline_outputs == ["test_output1.txt", "test,output2.txt"]
    assert '--outputs "test_output1.txt;test,output2.txt"' in notebook_op.container.args[0]


def test_construct_with_bad_pipeline_outputs():
    with pytest.raises(ValueError) as error_info:
        ExecuteFileOp(
            name="test",
            pipeline_name="test-pipeline",
            experiment_name="experiment-name",
            notebook="test_notebook.ipynb",
            cos_endpoint="http://testserver:32525",
            cos_bucket="test_bucket",
            cos_directory="test_directory",
            cos_dependencies_archive="test_archive.tgz",
            pipeline_outputs=["test_output1.txt", "test;output2.txt"],
            image="test/image:dev",
        )
    assert "Illegal character (;) found in filename 'test;output2.txt'." == str(error_info.value)


def test_construct_with_env_variables_argo():
    notebook_op = ExecuteFileOp(
        name="test",
        pipeline_name="test-pipeline",
        experiment_name="experiment-name",
        notebook="test_notebook.ipynb",
        cos_endpoint="http://testserver:32525",
        cos_bucket="test_bucket",
        cos_directory="test_directory",
        cos_dependencies_archive="test_archive.tgz",
        pipeline_envs={"ENV_VAR_ONE": "1", "ENV_VAR_TWO": "2", "ENV_VAR_THREE": "3"},
        image="test/image:dev",
    )

    confirmation_names = ["ENV_VAR_ONE", "ENV_VAR_TWO", "ENV_VAR_THREE", "ELYRA_RUN_NAME"]
    confirmation_values = ["1", "2", "3", RUN_ID_PLACEHOLDER]
    for env_val in notebook_op.container.env:
        assert env_val.name in confirmation_names
        assert env_val.value in confirmation_values
        confirmation_names.remove(env_val.name)
        confirmation_values.remove(env_val.value)

    # Verify confirmation values have been drained.
    assert len(confirmation_names) == 0
    assert len(confirmation_values) == 0

    # same as before but explicitly specify the workflow engine type
    # as Argo
    notebook_op = ExecuteFileOp(
        name="test",
        pipeline_name="test-pipeline",
        experiment_name="experiment-name",
        notebook="test_notebook.ipynb",
        cos_endpoint="http://testserver:32525",
        cos_bucket="test_bucket",
        cos_directory="test_directory",
        cos_dependencies_archive="test_archive.tgz",
        pipeline_envs={"ENV_VAR_ONE": "1", "ENV_VAR_TWO": "2", "ENV_VAR_THREE": "3"},
        image="test/image:dev",
        workflow_engine="Argo",
    )

    confirmation_names = ["ENV_VAR_ONE", "ENV_VAR_TWO", "ENV_VAR_THREE", "ELYRA_RUN_NAME"]
    confirmation_values = ["1", "2", "3", RUN_ID_PLACEHOLDER]
    for env_val in notebook_op.container.env:
        assert env_val.name in confirmation_names
        assert env_val.value in confirmation_values
        confirmation_names.remove(env_val.name)
        confirmation_values.remove(env_val.value)

    # Verify confirmation values have been drained.
    assert len(confirmation_names) == 0
    assert len(confirmation_values) == 0


def test_construct_with_env_variables_tekton():
    notebook_op = ExecuteFileOp(
        name="test",
        pipeline_name="test-pipeline",
        experiment_name="experiment-name",
        notebook="test_notebook.ipynb",
        cos_endpoint="http://testserver:32525",
        cos_bucket="test_bucket",
        cos_directory="test_directory",
        cos_dependencies_archive="test_archive.tgz",
        pipeline_envs={"ENV_VAR_ONE": "1", "ENV_VAR_TWO": "2", "ENV_VAR_THREE": "3"},
        image="test/image:dev",
        workflow_engine="Tekton",
    )

    confirmation_names = ["ENV_VAR_ONE", "ENV_VAR_TWO", "ENV_VAR_THREE", "ELYRA_RUN_NAME"]
    confirmation_values = ["1", "2", "3"]
    field_path = "metadata.annotations['pipelines.kubeflow.org/run_name']"
    for env_val in notebook_op.container.env:
        assert env_val.name in confirmation_names
        confirmation_names.remove(env_val.name)
        if env_val.name == "ELYRA_RUN_NAME":
            assert env_val.value_from.field_ref.field_path == field_path, env_val.value_from.field_ref
        else:
            assert env_val.value in confirmation_values
            confirmation_values.remove(env_val.value)

    # Verify confirmation values have been drained.
    assert len(confirmation_names) == 0
    assert len(confirmation_values) == 0


def test_normalize_label_value():
    valid_middle_chars = "-_."

    # test min length
    assert ExecuteFileOp._normalize_label_value(None) == ""
    assert ExecuteFileOp._normalize_label_value("") == ""
    # test max length (63)
    assert ExecuteFileOp._normalize_label_value("a" * 63) == "a" * 63
    assert ExecuteFileOp._normalize_label_value("a" * 64) == "a" * 63  # truncated
    # test first and last char
    assert ExecuteFileOp._normalize_label_value("1") == "1"
    assert ExecuteFileOp._normalize_label_value("22") == "22"
    assert ExecuteFileOp._normalize_label_value("3_3") == "3_3"
    assert ExecuteFileOp._normalize_label_value("4u4") == "4u4"
    assert ExecuteFileOp._normalize_label_value("5$5") == "5_5"

    # test first char
    for c in string.printable:
        if c in string.ascii_letters + string.digits:
            # first char is valid
            # no length violation
            assert ExecuteFileOp._normalize_label_value(c) == c
            assert ExecuteFileOp._normalize_label_value(c + "B") == c + "B"
            # max length
            assert ExecuteFileOp._normalize_label_value(c + "B" * 62) == (c + "B" * 62)
            # max length exceeded
            assert ExecuteFileOp._normalize_label_value(c + "B" * 63) == (c + "B" * 62)  # truncated
        else:
            # first char is invalid, e.g. '#a', and becomes the
            # second char, which might require replacement
            rv = c
            if c not in valid_middle_chars:
                rv = "_"
            # no length violation
            assert ExecuteFileOp._normalize_label_value(c) == "a" + rv + "a"
            assert ExecuteFileOp._normalize_label_value(c + "B") == "a" + rv + "B"
            # max length
            assert ExecuteFileOp._normalize_label_value(c + "B" * 62) == ("a" + rv + "B" * 61)  # truncated
            # max length exceeded
            assert ExecuteFileOp._normalize_label_value(c + "B" * 63) == ("a" + rv + "B" * 61)  # truncated

    # test last char
    for c in string.printable:
        if c in string.ascii_letters + string.digits:
            # no length violation
            assert ExecuteFileOp._normalize_label_value("b" + c) == "b" + c
            # max length
            assert ExecuteFileOp._normalize_label_value("b" * 62 + c) == ("b" * 62 + c)
            # max length exceeded
            assert ExecuteFileOp._normalize_label_value("b" * 63 + c) == ("b" * 63)
        else:
            # last char is invalid, e.g. 'a#', and requires
            # patching
            rv = c
            if c not in valid_middle_chars:
                rv = "_"
            # no length violation (char is appended)
            assert ExecuteFileOp._normalize_label_value("b" + c) == "b" + rv + "a"
            # max length (char is replaced)
            assert ExecuteFileOp._normalize_label_value("b" * 62 + c) == ("b" * 62 + "a")
            # max length exceeded (no action required)
            assert ExecuteFileOp._normalize_label_value("b" * 63 + c) == ("b" * 63)

    # test first and last char
    for c in string.printable:
        if c in string.ascii_letters + string.digits:
            # no length violation
            assert ExecuteFileOp._normalize_label_value(c + "b" + c) == c + "b" + c  # nothing is modified
            # max length
            assert ExecuteFileOp._normalize_label_value(c + "b" * 61 + c) == (c + "b" * 61 + c)  # nothing is modified
            # max length exceeded
            assert ExecuteFileOp._normalize_label_value(c + "b" * 62 + c) == c + "b" * 62  # truncate only
        else:
            # first and last characters are invalid, e.g. '#a#'
            rv = c
            if c not in valid_middle_chars:
                rv = "_"
            # no length violation
            assert ExecuteFileOp._normalize_label_value(c + "b" + c) == "a" + rv + "b" + rv + "a"
            # max length
            assert ExecuteFileOp._normalize_label_value(c + "b" * 59 + c) == ("a" + rv + "b" * 59 + rv + "a")
            # max length exceeded after processing, scenario 1
            # resolved by adding char before first, replace last
            assert ExecuteFileOp._normalize_label_value(c + "b" * 60 + c) == ("a" + rv + "b" * 60 + "a")
            # max length exceeded after processing, scenario 2
            # resolved by adding char before first, appending after last
            assert ExecuteFileOp._normalize_label_value(c + "b" * 59 + c) == ("a" + rv + "b" * 59 + rv + "a")
            # max length exceeded before processing, scenario 1
            # resolved by adding char before first, truncating last
            assert ExecuteFileOp._normalize_label_value(c + "b" * 62 + c) == ("a" + rv + "b" * 61)
            # max length exceeded before processing, scenario 2
            # resolved by adding char before first, replacing last
            assert ExecuteFileOp._normalize_label_value(c + "b" * 60 + c * 3) == ("a" + rv + "b" * 60 + "a")

    # test char in a position other than first and last
    # if invalid, the char is replaced with '_'
    for c in string.printable:
        if c in string.ascii_letters + string.digits + "-_.":
            assert ExecuteFileOp._normalize_label_value("A" + c + "Z") == "A" + c + "Z"
        else:
            assert ExecuteFileOp._normalize_label_value("A" + c + "Z") == "A_Z"

    # encore
    assert ExecuteFileOp._normalize_label_value(r"¯\_(ツ)_/¯") == "a_________a"



#
# Copyright 2018-2022 Elyra Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
import json
from logging import Logger
import os
from pathlib import Path
from queue import Empty
from queue import Queue
from threading import Event
from threading import Thread
import time
from typing import Dict
from typing import List
from typing import Optional
from typing import Union

import entrypoints
from jinja2 import Environment
from jinja2 import PackageLoader
from jinja2 import Template
from jupyter_core.paths import jupyter_runtime_dir
from traitlets.config import SingletonConfigurable
from watchdog.events import FileSystemEventHandler
from watchdog.observers import Observer

from elyra.metadata.manager import MetadataManager
from elyra.metadata.metadata import Metadata
from elyra.metadata.schemaspaces import ComponentCatalogs
from elyra.pipeline.catalog_connector import ComponentCatalogConnector
from elyra.pipeline.component import Component
from elyra.pipeline.component import ComponentParser
from elyra.pipeline.component_metadata import ComponentCatalogMetadata
from elyra.pipeline.runtime_type import RuntimeProcessorType

BLOCKING_TIMEOUT = 0.5
NONBLOCKING_TIMEOUT = 0.10
# Issue warnings if catalog update takes longer than this value in seconds
CATALOG_UPDATE_TIMEOUT = int(os.getenv("ELYRA_CATALOG_UPDATE_TIMEOUT", 15))
# Issue warnings when outstanding worker thread counts exceed this value
WORKER_THREAD_WARNING_THRESHOLD = int(os.getenv("ELYRA_WORKER_THREAD_WARNING_THRESHOLD", 10))


# Define custom type to describe the component cache
ComponentCacheType = Dict[str, Dict[str, Dict[str, Dict[str, Union[Component, str, List[str]]]]]]


class RefreshInProgressError(Exception):
    def __init__(self):
        super().__init__("A catalog refresh is in progress.  Try the request later.")


class RefreshQueue(Queue):
    """Entries are associated with a complete refresh of the Component Cache."""

    _refreshing: bool

    def __init__(self):
        super().__init__()
        self._refreshing = False

    @property
    def refreshing(self) -> bool:
        return self._refreshing

    @refreshing.setter
    def refreshing(self, value: bool) -> None:
        self._refreshing = value

    def get(self, block: bool = True, timeout: Optional[float] = None):
        """Overrides the superclass method to set the refreshing property to false when empty."""
        try:
            entry = super().get(block=block, timeout=timeout)
        except Empty:
            self.refreshing = False
            raise
        return entry

    def put(self, item, block=True, timeout=None):
        """Overrides the superclass method to set the refreshing property to true."""
        super().put(item, block=block, timeout=timeout)
        self.refreshing = True


class UpdateQueue(Queue):
    """Entries are associated with a single update of the Component Cache.

    This class merely exists to distinguish it from the RefreshQueue instance.
    """

    pass


class CacheUpdateManager(Thread):
    """
    Primary thread for maintaining consistency of the component cache.

    The component cache manager maintains the cache queue, whose entries are a
    tuple of 'catalog' and 'action'.  The 'catalog' is a catalog instance against
    which the 'action' is applied. The 'action' is one of 'modify' or 'delete'.
    For 'delete' the components of the referenced catalog are removed. For 'modify'
    the components of the referenced catalog are inserted or updated (depending on
    its prior existence).
    """

    def __init__(
        self, log: Logger, component_cache: ComponentCacheType, refresh_queue: RefreshQueue, update_queue: UpdateQueue
    ):
        super().__init__()

        self.daemon = True
        self.name = "CacheUpdateManager"

        self.log: Logger = log
        self._component_cache: ComponentCacheType = component_cache
        self._refresh_queue: RefreshQueue = refresh_queue
        self._update_queue: UpdateQueue = update_queue
        self._check_refresh_queue = False
        self._threads: List[CacheUpdateWorker] = []

        self.stop_event: Event = Event()  # Set when server process stops

    def run(self):
        """Process queue queue entries until server is stopped."""
        while not self.stop_event.is_set():
            self.manage_cache_tasks()

    def manage_cache_tasks(self):
        """
        Check the cache queue for a cache update action and start
        a corresponding worker thread to complete the update
        """
        outstanding_threads = self._has_outstanding_threads()
        try:
            # Get a task from the cache queue, waiting less if we have active threads.
            timeout = NONBLOCKING_TIMEOUT if outstanding_threads else BLOCKING_TIMEOUT

            # Toggle between refresh and update queues so as to prevent starvation.
            self._check_refresh_queue = not self._check_refresh_queue
            if self._check_refresh_queue:
                catalog, action = self._refresh_queue.get(timeout=timeout)
            else:
                catalog, action = self._update_queue.get(timeout=timeout)

        except Empty:
            # No task exists in the cache queue, proceed to check for thread execution
            pass

        else:
            # Create and start a thread for the task
            updater_thread = CacheUpdateWorker(
                self._component_cache,
                self._refresh_queue if self._check_refresh_queue else self._update_queue,
                catalog,
                action,
            )
            updater_thread.start()
            queue_clause = "refreshing" if self._check_refresh_queue else "updating"
            self.log.debug(f"CacheUpdateWorker {queue_clause} catalog: '{updater_thread.name}', action: '{action}'...")
            self._threads.append(updater_thread)

    def _has_outstanding_threads(self) -> bool:
        """
        Join finished threads and report on long-running threads as needed.
        """
        outstanding_threads = False
        for thread in self._threads:
            # Attempt to join thread within the given amount of time
            thread.join(timeout=NONBLOCKING_TIMEOUT)

            cumulative_run_time = int(time.time() - thread.task_start_time)
            if thread.is_alive():
                # Thread is still running (thread join timed out)
                outstanding_threads = True

                # Report on a long-running thread if CATALOG_UPDATE_TIMEOUT is exceeded
                time_since_last_check = int(time.time() - thread.last_warn_time)
                if time_since_last_check > CATALOG_UPDATE_TIMEOUT:
                    thread.last_warn_time = time.time()
                    self.log.warning(
                        f"Cache update for catalog '{thread.name}' is still processing "
                        f"after {cumulative_run_time} seconds ..."
                    )

            else:
                self.log.debug(f"CacheUpdateWorker completed for catalog: '{thread.name}', action: '{thread.action}'.")
                # Thread has been joined and can be removed from the list
                self._threads.remove(thread)

                # Mark cache task as complete
                thread.queue.task_done()

                # Report successful join for threads that have previously logged a
                # cache update duration warning
                if thread.last_warn_time != thread.task_start_time:
                    self.log.info(
                        f"Cache update for catalog '{thread.name}' has "
                        f"completed after {cumulative_run_time} seconds"
                    )

            if len(self._threads) > WORKER_THREAD_WARNING_THRESHOLD:
                self.log.warning(
                    f"CacheUpdateWorker outstanding threads threshold "
                    f"({WORKER_THREAD_WARNING_THRESHOLD}) has been exceeded. "
                    f"{len(self._threads)} threads are outstanding.  This may "
                    f"indicate a possible issue."
                )

        return outstanding_threads

    def is_refreshing(self) -> bool:
        return self._refresh_queue.refreshing

    def init_refresh(self) -> None:
        self._refresh_queue.refreshing = True

    def stop(self):
        """Trigger completion of the manager thread."""
        self._refresh_queue.refreshing = False
        self.stop_event.set()
        self.log.debug("CacheUpdateManager stopped.")


class CacheUpdateWorker(Thread):
    """Spawned by the CacheUpdateManager to perform work against the component cache."""

    def __init__(
        self,
        component_cache: ComponentCacheType,
        queue: Queue,
        catalog: ComponentCatalogMetadata,
        action: Optional[str] = None,
    ):

        super().__init__()

        self.daemon = True
        self.name = catalog.name  # Let the name of the thread reflect the catalog being managed

        self._component_cache: ComponentCacheType = component_cache

        # Task-specific properties
        self.queue: Queue = queue
        self.catalog: ComponentCatalogMetadata = catalog
        self.action: str = action

        # Thread metadata
        self.task_start_time = time.time()
        self.last_warn_time = self.task_start_time

        # Prepare component cache for modification
        runtime_type = None
        if self.catalog.metadata:
            runtime_type = self.catalog.runtime_type.name
        self.prepare_cache_for_catalog(runtime_type)

    def run(self):
        """Apply the relative action to the given catalog entry in the cache."""
        if self.action == "delete":
            # Check all runtime types in cache for an entry of the given name.
            # If found, remove only the components from this catalog
            for runtime_type in self._component_cache:
                if self.catalog.name in self._component_cache[runtime_type]:
                    self._component_cache[runtime_type].pop(self.catalog.name, None)
                    break
        else:  # 'modify' - replace (or add) components from the given catalog an update its status
            runtime_type = self.catalog.runtime_type.name
            catalog_state = self._component_cache[runtime_type][self.catalog.name].get("status")
            try:
                # Replace all components for the given catalog
                self._component_cache[runtime_type][self.catalog.name][
                    "components"
                ] = ComponentCache.instance().read_component_catalog(self.catalog)
                catalog_state["state"] = "current"
                catalog_state["errors"] = []  # reset any errors that may have been present
            except Exception as e:
                # Update state with an 'error' action and the relevant message
                catalog_state["state"] = "error"
                catalog_state["errors"].append(str(e))

    def prepare_cache_for_catalog(self, runtime_type: Optional[str] = None):
        """
        Add entries to the component cache for the runtime type and/or catalog
        of focus for this thread, and set the catalog state to 'updating'.
        """
        if self.action == "delete":
            # On 'delete' the runtime_type parameter will be None and since catalog names
            # are essentially unique across runtime types, we can break out of this loop
            # on first occurrence and let _that_ runtime type be used in the following code.
            for runtime_type in self._component_cache:
                if self.catalog.name in self._component_cache[runtime_type]:
                    break

        # Add sub-dictionary for this runtime type if not present
        if not self._component_cache.get(runtime_type):
            self._component_cache[runtime_type] = {}

        # Add sub-dictionary for this catalog if not present - this will occur when
        # a catalog instance is created, so we're essentially adding a placeholder.
        if not self._component_cache[runtime_type].get(self.catalog.name):
            self._component_cache[runtime_type][self.catalog.name] = {
                "components": {},
                "status": {"state": "updating", "errors": []},
            }
        else:  # Set state to 'updating' for an existing entry
            self._component_cache[runtime_type][self.catalog.name]["status"]["state"] = "updating"


class ComponentCache(SingletonConfigurable):
    """Represents the cache of component definitions indexed by runtime-type, then by catalog name."""

    # The component_cache is indexed at the top level by runtime type name, e.g. 'APACHE_AIRFLOW',
    # and has as its value another dictionary. At the second level, each sub-dictionary is indexed by
    # a ComponentCatalogMetadata instance name; its value is also a sub-dictionary. This sub-dictionary
    # consists of two additional dictionaries: 1.) one with key "components" whose dictionary is
    # indexed by component id and maps to the corresponding Component object, and 2.) one with key
    # "status" and value of a final sub-dictionary with key-value pairs "state":"<current/updating/errors>"
    # and "errors":["<error1>", "<error2>", ...] to dynamically indicate the status of this catalog instance
    _component_cache: ComponentCacheType = {}

    _generic_category_label = "Elyra"
    _generic_components: Dict[str, Component] = {
        "notebook": Component(
            id="notebook",
            name="Notebook",
            description="Run notebook file",
            op="execute-notebook-node",
            catalog_type="elyra",
            component_reference="elyra",
            extensions=[".ipynb"],
            categories=[_generic_category_label],
        ),
        "python-script": Component(
            id="python-script",
            name="Python Script",
            description="Run Python script",
            op="execute-python-node",
            catalog_type="elyra",
            component_reference="elyra",
            extensions=[".py"],
            categories=[_generic_category_label],
        ),
        "r-script": Component(
            id="r-script",
            name="R Script",
            description="Run R script",
            op="execute-r-node",
            catalog_type="elyra",
            component_reference="elyra",
            extensions=[".r"],
            categories=[_generic_category_label],
        ),
    }

    def __init__(self, **kwargs):
        super().__init__(**kwargs)

        self._component_cache = {}
        self.is_server_process = ComponentCache._determine_server_process(**kwargs)
        self.manifest_dir = jupyter_runtime_dir()
        # Ensure queue attribute exists for non-server instances as well.
        self.refresh_queue: Optional[RefreshQueue] = None
        self.update_queue: Optional[UpdateQueue] = None
        if self.is_server_process:
            self.refresh_queue = RefreshQueue()
            self.update_queue = UpdateQueue()

            # Set up watchdog for manifest file for out-of-process updates
            self.observer = Observer()
            self.observer.schedule(ManifestFileChangeHandler(self), self.manifest_dir)

            # Start a thread to manage updates to the component cache
            manager = CacheUpdateManager(self.log, self._component_cache, self.refresh_queue, self.update_queue)
            self.cache_manager = manager
            self.cache_manager.start()
            self.log.debug("CacheUpdateManager started...")
        else:
            self.manifest_filename = os.path.join(self.manifest_dir, f"elyra-component-manifest-{os.getpid()}.json")

    @staticmethod
    def _determine_server_process(**kwargs) -> bool:
        """Determines if this process is a server (extension) process."""
        app_names = ["ServerApp", "ElyraApp"]
        is_server_process = False
        if "parent" in kwargs and kwargs["parent"].__class__.__name__ in app_names:
            is_server_process = True
        elif "emulate_server_app" in kwargs and kwargs["emulate_server_app"]:  # Used in unittests
            is_server_process = True

        return is_server_process

    def load(self):
        """
        Completes a series of actions during system startup, such as creating
        the component manifest file and triggering the build of the component
        cache for existing ComponentCatalog metadata instances.
        """
        # Proceed only if singleton instance has been created
        if self.initialized:
            # The cache manager will work on manifest and cache tasks on an
            # in-process basis as load() is only called during startup from
            # the server process.
            if self.is_server_process:
                # Remove all existing manifest files from previous processes
                self._remove_all_manifest_files()

                # Start the watchdog if it's not alive, prevents redundant starts
                if not self.observer.is_alive():
                    self.observer.start()

                # Fetch all component catalog instances and trigger their add to the
                # component cache if this is not already happening (it seems some server
                # test fixtures could be loading the server extensions multiple times).
                if not self.cache_manager.is_refreshing():
                    self.refresh()

    def refresh(self):
        """Triggers a refresh of all catalogs in the component cache.

        Raises RefreshInProgressError if a complete refresh is in progress.
        Note that we do not preclude non-server processes from performing a
        complete refresh.  In such cases, each of the catalog entries will be
        written to the manifest, which will be placed into the update queue.
        As a result, non-server applications could by-pass the "refresh in progress"
        constraint, but we're assuming a CLI application won't be as likely to
        "pound" refresh like a UI application can.
        """
        if self.is_server_process and self.cache_manager.is_refreshing():
            raise RefreshInProgressError()
        catalogs = MetadataManager(schemaspace=ComponentCatalogs.COMPONENT_CATALOGS_SCHEMASPACE_ID).get_all()
        for catalog in catalogs:
            self._insert_request(self.refresh_queue, catalog, "modify")

    def update(self, catalog: Metadata, action: str):
        """
        Triggers an update of the component cache for the given catalog name.  If this is a non-server
        process, the entry is written to the manifest file where it will be "processed" by the watchdog
        and inserted into the component cache queue, otherwise we update the cache queue directly.
        """
        self._insert_request(self.update_queue, catalog, action)

    def _insert_request(self, queue: Queue, catalog: Metadata, action: str):
        """
        If running as a server process, the request is submitted to the desired queue, otherwise
        it is posted to the manifest where the server process (if running) can detect the manifest
        file update and send the request to the update queue.

        Note that any calls to ComponentCache.refresh() from non-server processes will still
        perform the refresh, but via the update queue rather than the refresh queue.  We could,
        instead, raise NotImplementedError in such cases, but we may want the ability to refresh
        the entire component cache from a CLI utility and the current implementation would allow that.
        """
        if self.is_server_process:
            queue.put((catalog, action))
        else:
            manifest: Dict[str, str] = self._load_manifest()
            manifest[catalog.name] = action
            self.update_manifest(manifest=manifest)

    def _remove_all_manifest_files(self):
        """
        Remove all existing manifest files in the Jupyter runtimes directory.
        """
        manifest_files = Path(self.manifest_dir).glob("**/elyra-component-manifest-*.json")
        for file in manifest_files:
            os.remove(str(file))

    def _load_manifest(self, filename: Optional[str] = None) -> Dict[str, str]:
        """Read and return the contents of a manifest file.

        If 'filename' is not provided, this process's manifest file will be read.
        """
        filename = filename or self.manifest_filename
        if not os.path.isfile(filename):
            self.log.debug(f"Manifest file '{filename}' doesn't exist and will be created.")
            return {}
        with open(filename, "r") as f:
            manifest: Dict[str, str] = json.load(f)
        self.log.debug(f"Reading manifest '{manifest}' from file '{filename}'")
        return manifest

    def update_manifest(self, filename: Optional[str] = None, manifest: Optional[Dict[str, str]] = None) -> None:
        """Update the manifest file with the given entry."""
        filename = filename or self.manifest_filename
        manifest = manifest or {}
        self.log.debug(f"Updating manifest '{manifest}' to file '{filename}'")
        with open(filename, "w") as f:
            json.dump(manifest, f, indent=2)

    def wait_for_all_cache_tasks(self):
        """
        Block execution and wait for all tasks in the cache task update queue to complete.
        Primarily used for testing.
        """
        if self.is_server_process:
            self.update_queue.join()
            self.refresh_queue.join()

    def get_all_components(self, platform: RuntimeProcessorType) -> List[Component]:
        """
        Retrieve all components from component catalog cache
        """
        components: List[Component] = []

        catalogs = self._component_cache.get(platform.name, {})
        for catalog_name, catalog_properties in catalogs.items():
            components.extend(list(catalog_properties.get("components", {}).values()))

        if not components and platform != RuntimeProcessorType.LOCAL:
            self.log.error(f"No components could be found in any catalog for platform type '{platform.name}'.")

        return components

    def get_component(self, platform: RuntimeProcessorType, component_id: str) -> Optional[Component]:
        """
        Retrieve the component with a given component_id from component catalog cache
        """
        component: Optional[Component] = None

        catalogs = self._component_cache.get(platform.name, {})
        for catalog_name, catalog_properties in catalogs.items():
            component = catalog_properties.get("components", {}).get(component_id)
            if component:
                break

        if not component:
            self.log.error(f"Component with ID '{component_id}' could not be found in any catalog.")

        return component

    def _load_catalog_reader_class(
        self, catalog: ComponentCatalogMetadata, file_types: List[str]
    ) -> Optional[ComponentCatalogConnector]:
        """
        Load the appropriate entrypoint class based on the schema name indicated in
        the ComponentCatalogMetadata instance and the file types associated with the component
        parser in use
        """
        try:
            catalog_reader = entrypoints.get_group_named("elyra.component.catalog_types").get(catalog.schema_name)
            if not catalog_reader:
                self.log.error(
                    f"No entrypoint with name '{catalog.schema_name}' was found in group "
                    f"'elyra.component.catalog_types' to match the 'schema_name' given in catalog "
                    f"'{catalog.display_name}'. Skipping..."
                )
                return None

            catalog_reader = catalog_reader.load()(file_types, parent=self.parent)
        except Exception as e:
            self.log.error(f"Could not load appropriate ComponentCatalogConnector class: {e}. Skipping...")
            return None

        return catalog_reader

    def read_component_catalog(self, catalog: ComponentCatalogMetadata) -> Dict[str, Component]:
        """
        Read a component catalog and return a dictionary of components indexed by component_id.

        :param catalog: a metadata instances from which to read and construct Component objects

        :returns: a dictionary of component id to Component object for all read/parsed components
        """
        components: Dict[str, Component] = {}

        # Assign component parser based on the runtime platform type
        parser = ComponentParser.create_instance(platform=catalog.runtime_type)

        # Assign reader based on the type of the catalog (the 'schema_name')
        catalog_reader = self._load_catalog_reader_class(catalog, parser.file_types)
        if not catalog_reader:
            return components

        # Get content of component definition file for each component in this catalog
        self.log.debug(f"Processing components in catalog '{catalog.display_name}'")
        catalog_entries = catalog_reader.read_component_definitions(catalog)
        if not catalog_entries:
            return components

        for catalog_entry in catalog_entries:
            # Parse the entry to get a fully qualified Component object
            try:
                parsed_components = parser.parse(catalog_entry) or []
            except Exception as e:
                self.log.warning(
                    f"Could not parse definition for component with identifying information: "
                    f"'{catalog_entry.entry_reference}' -> {str(e)}"
                )
            else:
                for component in parsed_components:
                    components[component.id] = component

        return components

    @staticmethod
    def get_generic_components() -> List[Component]:
        return list(ComponentCache._generic_components.values())

    @staticmethod
    def get_generic_component(component_id: str) -> Optional[Component]:
        return ComponentCache._generic_components.get(component_id)

    @staticmethod
    def get_generic_component_ops() -> List[str]:
        return [component.op for component in ComponentCache.get_generic_components()]

    @staticmethod
    def load_jinja_template(template_name: str) -> Template:
        """
        Loads the jinja template of the given name from the
        elyra/templates/components folder
        """
        loader = PackageLoader("elyra", "templates/components")
        template_env = Environment(loader=loader)

        return template_env.get_template(template_name)

    @staticmethod
    def to_canvas_palette(components: List[Component]) -> Dict:
        """
        Converts catalog components into appropriate canvas palette format
        """
        template = ComponentCache.load_jinja_template("canvas_palette_template.jinja2")

        # Define a fallback category for components with no given categories
        fallback_category_name = "No Category"

        # Convert the list of all components into a dictionary of
        # component lists keyed by category
        category_to_components: Dict[str, List[Component]] = {}
        for component in components:
            categories = component.categories

            # Assign a fallback category so that component is not
            # lost during palette render
            if not categories:
                categories = [fallback_category_name]

            for category in categories:
                if category not in category_to_components.keys():
                    category_to_components[category] = []

                if component.id not in [comp.id for comp in category_to_components[category]]:
                    category_to_components[category].append(component)

        # Render template
        canvas_palette = template.render(category_dict=category_to_components)
        return json.loads(canvas_palette)

    @staticmethod
    def to_canvas_properties(component: Component) -> Dict:
        """
        Converts catalog components into appropriate canvas properties format

        If component_id is one of the generic set, generic template is rendered,
        otherwise, the  runtime-specific property template is rendered
        """
        if ComponentCache.get_generic_component(component.id) is not None:
            template = ComponentCache.load_jinja_template("generic_properties_template.jinja2")
        else:
            template = ComponentCache.load_jinja_template("canvas_properties_template.jinja2")

        canvas_properties = template.render(component=component)
        return json.loads(canvas_properties)


class ManifestFileChangeHandler(FileSystemEventHandler):
    """Watchdog handler that filters on .json files within specific metadata directories."""

    def __init__(self, component_cache: ComponentCache, **kwargs):
        super().__init__(**kwargs)
        self.component_cache = component_cache
        self.log = component_cache.log

    def dispatch(self, event):
        """Dispatches delete and modification events pertaining to the manifest filename"""
        if "elyra-component-manifest" in event.src_path:
            super().dispatch(event)

    def on_modified(self, event):
        """Fires when the component manifest file is modified."""
        self.log.debug(f"ManifestFileChangeHandler: file '{event.src_path}' has been modified.")
        manifest = self.component_cache._load_manifest(filename=event.src_path)
        if manifest:  # only update the manifest if there is work to do
            for catalog, action in manifest.items():
                self.log.debug(f"ManifestFileChangeHandler: inserting ({catalog},{action}) into update queue...")
                if action == "delete":
                    # The metadata instance has already been deleted, so we must
                    # fabricate an instance that only consists of a catalog name
                    catalog_instance = ComponentCatalogMetadata(name=catalog)

                else:  # cache_action == 'modify':
                    # Fetch the catalog instance associated with this action
                    catalog_instance = MetadataManager(
                        schemaspace=ComponentCatalogs.COMPONENT_CATALOGS_SCHEMASPACE_ID
                    ).get(name=catalog)

                self.component_cache.update(catalog=catalog_instance, action=action)
            self.component_cache.update_manifest(filename=event.src_path)  # clear the manifest



#
# Copyright 2018-2022 Elyra Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Path utilities"""

import os


def get_absolute_path(root_dir: str, path: str) -> str:
    """Checks if path is absolute or not.  If not absolute, `path` is appended to `root_dir`."""

    absolute_path = os.path.normpath(path)
    if not os.path.isabs(absolute_path):
        absolute_path = os.path.join(os.path.normpath(root_dir), absolute_path)

    return absolute_path


def get_expanded_path(root_dir: str = None) -> str:
    # Since root_dir may contain '~' for user home, use expanduser()
    return os.path.expanduser(root_dir or os.getcwd())
